from __future__ import absolute_import
'''Module providing a data translation methods'''
from KDEpy import FFTKDE
from collections.abc import Iterable
import numpy as np
from dama.translations import Translation

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class KDE(Translation):
    def __init__(
        self,
        source,
        *args,
        bw=None,
        kernel='gaussian',
        density=True,
        **kwargs
        ):
        '''run KDE on regular grid

        Parameters:
        -----------

        source : GridData or PointData
        bw : str or float or iterable
            Will default to 'silverman; for 1d data and 1 otherwise
            coices of 'silverman', 'scott', 'ISJ' for 1d data
            float specifies fixed bandwidth, in case of iterable a separate, fixed bandwith per dimension
        kernel : str
            choices of 'gaussian', 'exponential', 'box', 'tri', 'epa', 'biweight', 'triweight', 'tricube', 'cosine'
        density : bool (optional)
            if false, multiply output by sum of data
        '''
        super().__init__(source, *args, dest_needs_grid=True, **kwargs)

        if bw is None:
            bw = 'silverman' if self.dest.grid.nax == 1 else 1.
        self.bw = bw
        self.kernel = kernel
        self.density = density

        if density:
            self.additional_runs = {'density': None}
        else:
            self.additional_runs = {'counts': None}

        if not self.dest.grid.regular:
            raise TypeError('dest must have regular grid')

    def setup(self):
        self.prepare_source_sample(stacked=False)
        # every point must be inside output grid (requirement of KDEpy)
        masks = [
            np.logical_and(
                self.source_sample[i] > dim.points[0],
                self.source_sample[i] < dim.points[-1]
                ) for i, dim in enumerate(self.dest.grid)
            ]
        self.mask = np.all(masks, axis=0)
        #n_masked = np.sum(~mask)
        #if n_masked > 0:
        #    warnings.warn('Excluding %i points that are outside grid'%n_masked, Warning, stacklevel=0)
        sample = [s[self.mask] for s in self.source_sample]

        self.source_sample = np.stack(sample).T
        self.prepare_dest_sample(transposed=True)

        if isinstance(self.bw, (np.ndarray, list, tuple)):
            for i in range(self.dest.grid.nax):
                self.source_sample[:, i] /= self.bw[i]
                self.dest_sample[:, i] /= self.bw[i]
            bw = 1
        else:
            bw = self.bw

        self.kde = FFTKDE(bw=bw, kernel=self.kernel)

    def eval(self, source_data):

        if self.density:
            # since we scale the inputs, we need to re-scale the
            # densities such that they integrate out to being 1 again
            if isinstance(self.bw, (np.ndarray, list, tuple)):
                scale = 1. / np.prod(self.bw)
            else:
                scale = None

        if source_data is None:
            out_array = self.kde.fit(self.source_sample
                                     ).evaluate(self.dest_sample)
            out_shape = self.dest.shape
            if not self.density:
                out_array *= self.source_sample.size / np.sum(out_array)
            elif scale is not None:
                out_array *= scale

        else:
            source_data = source_data.flat()

            if source_data.ndim > 1:
                out_array = self.get_empty_output_array(
                    source_data.shape[1:], flat=True
                    )
                for idx in np.ndindex(*source_data.shape[1:]):
                    out_array[(Ellipsis, ) + idx] = self.kde.fit(
                        self.source_sample,
                        weights=source_data[(Ellipsis, ) + idx][self.mask]
                        ).evaluate(self.dest_sample)
                    if not self.density:
                        out_array[(Ellipsis, ) + idx] *= np.sum(
                            source_data[(Ellipsis, ) + idx][self.mask]
                            ) / np.sum(out_array[(Ellipsis, ) + idx])
                    elif scale is not None:
                        out_array[(Ellipsis, ) + idx] *= scale
                out_shape = (self.dest.shape) + (-1, )

            else:
                out_array = self.kde.fit(
                    self.source_sample, weights=source_data[self.mask]
                    ).evaluate(self.dest_sample)
                out_shape = self.dest.shape
                if not self.density:
                    out_array *= np.sum(source_data[self.mask]
                                        ) / np.sum(out_array)
                elif scale is not None:
                    out_array *= scale

        #if isinstance(self.bw, (np.ndarray, list, tuple)):
        #    out_array *= np.product(self.bw)

        return out_array.reshape(out_shape)



from __future__ import absolute_import
'''Module providing a data translation methods'''
import numpy as np
from scipy import interpolate

from dama.translations import Translation

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class Interpolation(Translation):
    def __init__(
        self, source, *args, method=None, fill_value=np.nan, **kwargs
        ):
        '''interpolation from any data to another

        Parameters:
        -----------
        method : string
            "nearest" = nearest neightbour interpolation
            "linear" = linear interpolation
            "cubic" = cubic interpolation (only for ndim < 3)
        fill_value : Number (optional)
            value for invalid points
        '''
        super().__init__(source, *args, **kwargs)

        assert method in [
            None, 'nearest', 'linear', 'cubic'
            ], 'Illegal method %s' % method

        if method is None:
            if len(self.wrt) > 2:
                method = 'linear'
            else:
                method = 'cubic'
        self.method = method
        self.fill_value = fill_value

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):

        if self.dest_has_grid:
            source_data = source_data.flat()
            mask = np.isfinite(source_data)
            if mask.ndim > 1:
                dim_mask = np.all(mask, axis=1)
            else:
                dim_mask = mask

            sample = self.source_sample[..., dim_mask]
            if source_data.ndim > 1:
                output_array = self.get_empty_output_array(
                    source_data.shape[1:]
                    )
                for idx in np.ndindex(*source_data.shape[1:]):
                    output_array[(Ellipsis, ) + idx] = interpolate.griddata(
                        points=sample.T,
                        values=source_data[(Ellipsis, ) + idx][dim_mask],
                        xi=self.dest_sample.T,
                        method=self.method,
                        fill_value=self.fill_value
                        ).T
            else:
                output_array = interpolate.griddata(
                    points=sample.T,
                    values=source_data[mask],
                    xi=self.dest_sample.T,
                    method=self.method,
                    fill_value=self.fill_value
                    ).T
                output_array = np.squeeze(output_array)
            return output_array

        else:
            if len(self.wrt) == 1:
                f = interpolate.interp1d(
                    self.source_sample[0],
                    source_data,
                    kind=self.method,
                    fill_value=self.fill_value,
                    bounds_error=False
                    )
                return f(self.dest_sample[0])

            elif len(self.wrt) == 2:
                f = interpolate.interp2d(
                    self.source_sample[0],
                    self.source_sample[1],
                    source_data,
                    kind=self.method,
                    fill_value=self.fill_value,
                    bounds_error=False
                    )
                return np.array(
                    [
                        f(x, y)[0] for x, y in
                        zip(self.dest_sample[0], self.dest_sample[1])
                        ]
                    )

            else:
                if self.method == 'nearest':
                    f = interpolate.NearestNDInterpolator(
                        self.source_sample.T, source_data
                        )
                else:
                    f = interpolate.LinearNDInterpolator(
                        self.source_sample.T,
                        source_data,
                        fill_value=self.fill_value
                        )
                return f(self.dest_sample.T)



from __future__ import absolute_import
import numpy as np

try:
    import pandas
except ImportError:
    pandas = None

import dama as dm
from dama import translations
import dama.plotting
from dama.utils.formatter import format_table

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class PointData(object):
    '''
    Data Layer to hold point-type data structures (Pandas DataFrame, Dict, )
    '''
    __slots__ = [
            '_data',
            ]

    def __init__(self, *args, **kwargs):
        self._data = {}

        if len(args) == 0 and len(kwargs) == 0:
            pass
        elif len(args) == 1 and len(kwargs) == 0:
            if pandas and isinstance(args[0], pandas.core.series.Series):
                self._data = pandas.DataFrame(args[0])
            elif pandas and isinstance(args[0], pandas.core.frame.DataFrame):
                self._data = args[0]
            elif isinstance(args[0], dict):
                kwargs = args[0]
            else:
                raise ValueError("Did not understand input arguments")
        if len(kwargs) > 0:
            for k, v in kwargs.items():
                self[k] = v
        #else:
        #    raise ValueError("Did not understand input arguments")

    def __getstate__(self):
        return self._data

    def __setstate__(self, state):
        self._data = state

    def __repr__(self):
        return format_table(self, tablefmt='plain')

    def __str__(self):
        return format_table(self, tablefmt='plain')

    def _repr_html_(self):
        '''for jupyter'''
        if self.type == 'df':
            return self._data._repr_html_()
        return format_table(self, tablefmt='html')

    @property
    def data(self):
        return self._data

    @property
    def vars(self):
        '''
        Available variables
        '''
        if self.type == 'df':
            return list(self._data.columns)
        elif self.type == 'simple':
            return list(self._data.keys())
        return []

    def __dir__(self):
        """for tab completion"""
        return self.vars + object.__dir__(self)

    @property
    def data_vars(self):
        '''Available variables'''
        return self.vars

    @property
    def type(self):
        '''type of stored data

        Returns
        -------
        type : str
            if data is a pandas.DataFrame: "df"
            else: "simple"
        '''
        if pandas is not None:
            if isinstance(self._data, pandas.core.frame.DataFrame):
                return 'df'
        return 'simple'

    def __len__(self):
        return len(self._data)

    @property
    def size(self):
        return self.array_shape[0]

    @property
    def array_shape(self):
        '''the shape (first dimesnion only) of a single variable

        Returns
        -------
        shape : tuple
        '''
        if self.type == 'df':
            return (len(self._data), )
        elif self.type == 'simple':
            return (len(self[self.vars[0]]), )

    def __setitem__(self, var, val):
        if callable(val):
            self._data[var] = val
            return
        val = np.asanyarray(val)
        if val.ndim == 0:
            val = val[np.newaxis]

        if len(self) > 0:
            assert len(val) == self.array_shape[0], 'Incompatible dimensions'

        if isinstance(val, dm.PointArray):
            self._data[var] = val
        elif isinstance(val, np.ndarray):
            if self.type == 'df':
                self._data[var] = val
            else:
                val = dm.PointArray(val)
                self[var] = val
        else:
            raise ValueError()

    def __getitem__(self, var):
        if isinstance(var, str):
            if var in self.vars:
                data = self._data[var]
            else:
                raise IndexError('No variable %s in DataSet' % var)
            if callable(data):
                self[var] = data()
                data = self._data[var]
            if self.type == 'df':
                if isinstance(data, pandas.core.frame.DataFrame):
                    return dm.PointData(data)
                elif isinstance(data, pandas.core.series.Series):
                    return dm.PointArray(data)
            return data

        # create new instance with mask or slice applied
        new_data = {}

        if isinstance(var, (tuple,
                            list)) and all([isinstance(v, str) for v in var]):
            for v in var:
                new_data[v] = self[v]
        else:
            for n, d in self.items():
                new_data[n] = d[var]
        return dm.PointData(new_data)

    def __getattr__(self, item):
        try:
            return self[item]
        except Exception as e:
            raise AttributeError from e

    def __setattr__(self, item, value):
        if item in self.__slots__:
            object.__setattr__(self, item, value)
        else:
            self[item] = value

    def get_array(self, var, flat=False):
        return np.asarray(self[var])

    def __iter__(self):
        '''
        iterate over dimensions
        '''
        return iter([self[var] for var in self.vars])

    def items(self):
        return iter([(var, self[var]) for var in self.vars])

    # --- Tranlsation methods ---

    def interp(self, *args, method=None, fill_value=np.nan, **kwargs):
        return translations.Interpolation(
            self, *args, method=method, fill_value=fill_value, **kwargs
            ).run()

    interp.__doc__ = translations.Interpolation.__init__.__doc__

    def histogram(self, *args, density=False, **kwargs):
        return translations.Histogram(self, *args, density=density,
                                      **kwargs).run()

    histogram.__doc__ = translations.Histogram.__init__.__doc__

    def binwise(self, *args, **kwargs):
        return dm.BinnedData(data=self, *args, **kwargs)

    def kde(self, *args, **kwargs):
        return translations.KDE(self, *args, **kwargs).run()

    kde.__doc__ = translations.KDE.__init__.__doc__

    # --- Plotting functions ---

    plot = dm.plotting.plot

    def plot_2d(self, *args, labels=None, **kwargs):
        if len(args) == 2:
            dm.plotting.plot(self, *args, labels=labels, **kwargs)
        elif len(self) == 2:
            dm.plotting.plot(self, *self.vars, *args, **kwargs)
        else:
            raise ValueError('Need to specify 2 variables to plot')

    def plot_scatter(
        self, x, y, c=None, s=None, cbar=False, fig=None, ax=None, **kwargs
        ):
        dm.plotting.plot_points_2d(
            self, x, y, c=c, s=s, cbar=cbar, fig=fig, ax=ax, **kwargs
            )



from .translation import Translation
from .lookup import Lookup
from .interpolation import Interpolation
from .histogram import Histogram
from .kde import KDE
from .resample import Resample
from .ndimage import *



from __future__ import absolute_import
import numpy as np
import dama as dm
from dama.utils.formatter import format_table

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class PointArray(np.ndarray):
    '''Structure to hold a single PointData item
    '''
    def __new__(cls, input_array, *args, **kwargs):
        obj = np.asarray(input_array).view(cls)
        return obj

    def _repr_html_(self):
        '''for jupyter'''
        return format_table(self, tablefmt='html')

    def __repr__(self):
        return format_table(self, tablefmt='plain')

    def __str__(self):
        return format_table(self, tablefmt='plain')

    def __array_wrap__(self, obj):
        if obj.shape == ():
            return obj[()]  # if ufunc output is scalar, return it
        else:
            return np.ndarray.__array_wrap__(self, obj)

    def flat(self):
        return self

    def plot(self, *args, **kwargs):
        return dm.plotting.plot1d_all(self, *args, **kwargs)



import numpy as np
import matplotlib
import glob
import os

maps = glob.glob(os.path.join(os.path.dirname(__file__),'*.npy'))
cms = {}

for m in maps:
    name = os.path.splitext(os.path.basename(m))[0]
    array = np.load(m)
    cms[name] = matplotlib.colors.ListedColormap(array, name=name)
    cms[name+'_r'] = matplotlib.colors.ListedColormap(array[::-1], name=name+'_r')

locals().update(cms)



import os
try:
   import cPickle as pickle
except:
   import pickle


def save(filename, obj):
    basename, extension = os.path.splitext(filename)
    if extension == "":
        extension = ".pkl"
    if not extension == ".pkl":
        raise ValueError("Only pickle (.pkl) format supprted, but %s was provided"%extension)
    filename = basename + extension

    with open(filename, 'wb') as f:
        pickle.dump(obj, f)

def load(filename):
    basename, extension = os.path.splitext(filename)
    if extension == "":
        extension = ".pkl"
    if not extension == ".pkl":
        raise ValueError("Only pickle (.pkl) format supprted, but %s was provided"%extension)
    filename = basename + extension

    with open(filename, 'rb') as f:
        obj = pickle.load(f)
    return obj



from __future__ import absolute_import
from collections.abc import Iterable
import numpy as np
import numpy_indexed as npi
import dama as dm
from dama import translations
from dama.utils.formatter import format_table
import dama.plotting

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class BinnedData:
    '''
    Class to hold binned data
    '''

    # ToDo fill_value kwarg

    def __init__(self, grid=None, data=None, *args, **kwargs):
        '''
        grid : dm.Grid

        data : dm.PointData, dm.GridData or dm.GridArray
            -> if PointArray, sample is needed

        '''
        # set up grid
        if grid is None:
            if len(args) == 1 and len(kwargs
                                      ) == 0 and isinstance(args[0], dm.Grid):
                self.grid = args[0]
            else:
                self.grid = dm.Grid(*args, **kwargs)

        #assert set(self.grid.vars) <= set(data.vars)

        if not self.grid.initialized:
            self.grid.initialize(data)

        # create sample (grid variables) and remove from data
        self.sample = [
            data.get_array(var, flat=True) for var in self.grid.vars
            ]

        # The following is a bit wonky, ToDo
        if isinstance(data, dm.GridArray):
            self.single = True
            self.data = dm.PointData(test=data.flat())

        else:
            self.single = False
            data_vars = data.vars
            for var in self.grid.vars:
                data_vars.remove(var)

            if isinstance(data, dm.PointData):
                self.data = data[data_vars]

            elif isinstance(data, dm.GridData):
                self.data = dm.PointData()
                for var in data_vars:
                    self.data[var] = data.get_array(var)

            else:
                raise ValueError()

        self.indices = None
        self.group = None

    def compute_indices(self):
        self.indices = self.grid.compute_indices(self.sample)
        self.group = npi.group_by(self.indices)

    def __setitem__(self, var, val):
        self.add_data(var, val)

    def __getitem__(self, item):

        if isinstance(item, str):
            item = [item]

        item += self.grid.vars

        return dm.BinnedData(grid=self.grid, data=self.data[item])

    @property
    def vars(self):
        '''
        Available variables
        '''
        return self.grid.vars + self.data_vars

    @property
    def data_vars(self):
        '''
        only data variables (no grid vars)
        '''
        return list(self.data.keys())

    @property
    def shape(self):
        return self.grid.shape

    @property
    def ndim(self):
        return self.grid.nax

    @property
    def array_shape(self):
        '''
        shape of a single variable
        '''
        return self.shape

    def add_data(self, var, data):
        '''Add data

        Parameters
        ----------
        var : str
            name of data
        data : PointArray, PointData, Array
        '''
        if self.single:
            raise Exception('Cannot add data to a singular variable structure')

        if isinstance(data, dm.PointData):
            assert len(data.vars) == 1
            data = data[data.vars[0]]

        self.data[var] = data

    def run_np_indexed(self, method, fill_value=np.nan, **kwargs):
        '''run the numpy indexed methods
        Parameters:
        -----------

        method : str
            choice of ['sum', 'mean', 'min', 'max', 'std', 'var', 'argmin', 'argmax', 'median', 'mode', 'prod']
        '''
        axis = kwargs.pop('axis', None)
        assert axis is None, 'Axis kw not supported for BinnedData'

        if self.indices is None:
            self.compute_indices()
        outputs = {}
        output_maps = {}
        for var in self.data.vars:
            source_data = self.data[var]
            if source_data.ndim > 1:
                output_maps[var] = np.full(
                    self.grid.shape + source_data.shape[1:],
                    fill_value=fill_value,
                    dtype=self.data[var].dtype
                    )
            else:
                output_maps[var] = np.full(
                    self.grid.shape,
                    fill_value=fill_value,
                    dtype=self.data[var].dtype
                    )
            source_data = self.data[var]
            indices, outputs[var] = self.group.__getattribute__(method
                                                                )(source_data)

        for i, idx in enumerate(indices):
            if idx < 0:
                continue
            out_idx = np.unravel_index(idx, self.grid.shape)
            for var in self.data.vars:
                output_maps[var][out_idx] = result = outputs[var][i]

        # Pack into GridArray
        if self.single:
            out_data = dm.GridArray(output_maps['test'], grid=self.grid)

        else:
            # Pack into GridData
            out_data = dm.GridData(self.grid)
            for var, output_map in output_maps.items():
                out_data[var] = output_map

        return out_data

    def apply_function(
        self, function, *args, fill_value=np.nan, return_len=None, **kwargs
        ):
        '''apply function per bin

        function : callable
        return_len : int (optional)
            the shape (array length) of the `function` output. If None is passed, 
            `return_len` is (tried to be) determined on the fly

        if weights are passed as kwarg, they are interpreted as being a variable in the dataset
        '''

        if self.indices is None:
            self.compute_indices()

        outputs = {}
        output_maps = {}

        weights = kwargs.pop('weights', None)

        for var in self.data.vars:
            source_data = self.data[var]

            if return_len is None:
                # try to figure out return length of function

                if source_data.ndim > 1:
                    if weights is not None:
                        kwargs['weights'] = self.data[weights][:3, [0] * (source_data.ndim - 1)]
                    test_value = function(
                        source_data[:3, [0] * (source_data.ndim - 1)], *args,
                        **kwargs
                        )
                else:
                    if weights is not None:
                        kwargs['weights'] = self.data[weights][:3]
                    test_value = function(source_data[:3], *args, **kwargs)
                if np.isscalar(test_value):
                    return_len = 1
                else:
                    return_len = len(test_value)

            if source_data.ndim > 1:
                if return_len > 1:
                    output_maps[var] = np.full(
                        self.grid.shape + source_data.shape[1:] +
                        (return_len, ),
                        fill_value=fill_value,
                        dtype=source_data.dtype
                        )
                else:
                    output_maps[var] = np.full(
                        self.grid.shape + source_data.shape[1:],
                        fill_value=fill_value,
                        dtype=source_data.dtype
                        )
            else:
                if return_len > 1:
                    output_maps[var] = np.full(
                        self.grid.shape + (return_len, ),
                        fill_value=fill_value,
                        dtype=source_data.dtype
                        )
                else:
                    output_maps[var] = np.full(
                        self.grid.shape,
                        fill_value=fill_value,
                        dtype=source_data.dtype
                        )

        for i in range(self.grid.size):
            mask = self.indices == i
            if np.any(mask):
                out_idx = np.unravel_index(i, self.grid.shape)
                for var in self.data.vars:
                    source_data = self.data[var]

                    if source_data.ndim > 1:
                        for idx in np.ndindex(*source_data.shape[1:]):
                            if weights is not None:
                                kwargs['weights'] = self.data[weights][:, idx][mask]

                            output_maps[var][out_idx + (idx, )] = function(
                                source_data[:, idx][mask], *args, **kwargs
                                )
                    else:
                        if weights is not None:
                            kwargs['weights'] = self.data[weights][mask]

                        output_maps[var][out_idx] = function(
                            source_data[mask], *args, **kwargs
                            )

        # Pack into GridArray
        if self.single:
            out_data = dm.GridArray(output_maps['test'], grid=self.grid)

        # Pack into GridData
        else:
            out_data = dm.GridData(self.grid)
            for var, output_map in output_maps.items():
                out_data[var] = output_map

        return out_data

    def sum(self, fill_value=np.nan, **kwargs):
        return self.run_np_indexed('sum', fill_value=fill_value, **kwargs)

    def mean(self, fill_value=np.nan, **kwargs):
        return self.run_np_indexed('mean', fill_value=fill_value, **kwargs)

    def min(self, fill_value=np.nan, **kwargs):
        return self.run_np_indexed('min', fill_value=fill_value, **kwargs)

    def max(self, fill_value=np.nan, **kwargs):
        return self.run_np_indexed('max', fill_value=fill_value, **kwargs)

    def std(self, fill_value=np.nan, **kwargs):
        return self.run_np_indexed('std', fill_value=fill_value, **kwargs)

    def var(self, fill_value=np.nan, **kwargs):
        return self.run_np_indexed('var', fill_value=fill_value, **kwargs)

    def argmin(self, fill_value=np.nan, **kwargs):
        return self.run_np_indexed('argmin', fill_value=fill_value, **kwargs)

    def argmax(self, fill_value=np.nan, **kwargs):
        return self.run_np_indexed('argmax', fill_value=fill_value, **kwargs)

    def median(self, fill_value=np.nan, **kwargs):
        return self.run_np_indexed('median', fill_value=fill_value, **kwargs)

    def mode(self, fill_value=np.nan, **kwargs):
        return self.run_np_indexed('mode', fill_value=fill_value, **kwargs)

    def prod(self, fill_value=np.nan, **kwargs):
        return self.run_np_indexed('prod', fill_value=fill_value, **kwargs)

    def quantile(self, q, fill_value=np.nan, **kwargs):
        if not isinstance(q, Iterable):
            return_len = 1
        else:
            return_len = len(q)
        return self.apply_function(
            np.quantile,
            q,
            return_len=return_len,
            fill_value=fill_value,
            **kwargs
            )

    def std_devs(self, sigmas=[1.,], fill_value=np.nan, **kwargs):
        """ Mean + Sigma values
        sigmas : iterable
            values to produce bands = mean =/- sigmas * std
        """
        mean = self.mean(fill_value=np.nan, **kwargs)
        std = self.std(fill_value=np.nan, **kwargs)

        output_maps = {}

        for var in self.data.vars:
            m = np.ma.asarray(mean[var])
            s = np.ma.asarray(std[var])

            arrays = [m]
            for sigma in sigmas:
                delta = sigma * s
                arrays.append(m + delta)
                arrays.insert(0, m - delta)
        
            output_maps[var] = np.swapaxes(np.stack(arrays), 0, -1)
            

        # Pack into GridArray
        if self.single:
            out_data = dm.GridArray(output_maps['test'], grid=self.grid)

        # Pack into GridData
        else:
            out_data = dm.GridData(self.grid)
            for var, output_map in output_maps.items():
                out_data[var] = output_map

        return out_data



"""Function to bin methods
From: https://stackoverflow.com/questions/1015307/python-bind-an-unbound-method
"""


def bind(instance, func, as_name=None):
    """
    Bind the function *func* to *instance*, with either provided name *as_name*
    or the existing name of *func*. The provided *func* should accept the 
    instance as the first argument, i.e. "self".
    """
    if as_name is None:
        as_name = func.__name__
    bound_method = func.__get__(instance, instance.__class__)
    setattr(instance, as_name, bound_method)
    return bound_method



from __future__ import absolute_import
from collections import OrderedDict
from collections.abc import Iterable
import numpy as np
import dama as dm
from dama import translations
from dama.utils.formatter import format_table
from dama.utils.bind import bind
import dama.plotting

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class GridData:
    '''
    Class to hold grid data
    '''
    __slots__ = [
            '_data',
            '_grid',
            ]

    def __init__(self, *args, **kwargs):
        '''
        Set the grid
        '''
        self._data = {}

        self._grid = dm.Grid()

        if len(args) == 0 and len(kwargs) > 0 and all(
            [isinstance(v, dm.GridArray) for v in kwargs.values()]
            ):
            for n, d in kwargs.items():
                self.add_data(n, d)
        elif len(args) == 1 and len(kwargs
                                    ) == 0 and isinstance(args[0], dm.Grid):
            self._grid = args[0]
        else:
            self._grid = dm.Grid(*args, **kwargs)

    def __repr__(self):
        return format_table(self, tablefmt='plain')

    def __str__(self):
        return format_table(self, tablefmt='plain')

    def _repr_html_(self):
        '''for jupyter'''
        return format_table(self, tablefmt='html')

    def __getstate__(self):
        return self._data, self._grid

    def __setstate__(self, state):
        self._data, self._grid = state

    def __setitem__(self, var, val):
        self.add_data(var, val)

    def __getitem__(self, item):
        if isinstance(item, str):
            if item in self.vars:
                if item in self.data_vars:
                    data = self._data[item]
                    if callable(data):
                        self[item] = data()
                        data = self._data[item]
                else:
                    data = self.get_array(item)
                new_data = dm.GridArray(data, grid=self._grid)
                return new_data
            else:
                raise IndexError('No variable %s in DataSet' % item)

        # mask
        if isinstance(item, dm.GridArray):
            if item.dtype == np.bool:
                # in this case it is a mask
                # ToDo: masked operations behave strangely, operations are applyed to all elements, even if masked
                new_data = dm.GridData(self._grid)
                for v in self.data_vars:
                    new_data[v] = self[v][item]
                return new_data
            raise NotImplementedError('get item %s' % item)

        # create new instance with only those vars
        if isinstance(item,
                      Iterable) and all([isinstance(v, str) for v in item]):
            new_data = dm.GridData(self._grid)
            for v in item:
                if v in self.data_vars:
                    new_data[v] = self[v]
            return new_data

        # slice
        new_grid = self._grid[item]
        if len(new_grid) == 0:
            return {n: d[item] for n, d in self.items()}
        new_data = dm.GridData(new_grid)
        for n, d in self.items():
            new_data[n] = d[item]
        return new_data

    def __getattr__(self, item):
        try:
            return self[item]
        except Exception as e:
            raise AttributeError from e

    def __setattr__(self, item, value):
        if item in self.__slots__:
            object.__setattr__(self, item, value)
        else:
            self[item] = value

    @property
    def T(self):
        '''transpose'''
        new_obj = dm.GridData()
        new_obj._grid = self._grid.T
        for n, d in self.items():
            new_obj[n] = d.T
        return new_obj
        #raise NotImplementedError()

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        '''interface to numpy unversal functions'''
        scalar_results = OrderedDict()
        array_result = dm.GridData()
        for var in inputs[0].data_vars:
            converted_inputs = [inp[var] for inp in inputs]
            result = converted_inputs[0].__array_ufunc__(
                ufunc, method, *converted_inputs, **kwargs
                )
            if isinstance(result, dm.GridArray):
                array_result[var] = result
            else:
                scalar_results[var] = result
        if len(array_result.data_vars) == 0:
            return scalar_results
        if len(scalar_results) == 0:
            return array_result
        return scalar_results, array_result

    @property
    def vars(self):
        '''
        Available variables
        '''
        return self._grid.vars + self.data_vars

    def __dir__(self):
        """for tab completion"""
        return self.vars + object.__dir__(self)

    @property
    def data_vars(self):
        '''
        only data variables (no grid vars)
        '''
        return list(self._data.keys())

    @property
    def shape(self):
        return self._grid.shape

    @property
    def ndim(self):
        return self._grid.nax

    @property
    def array_shape(self):
        '''
        shape of a single variable
        '''
        return self.shape

    @property
    def grid(self):
        return self._grid

    @property
    def data(self):
        return self._data

    def add_data(self, var, data):
        '''Add data

        Parameters
        ----------
        var : str
            name of data
        data : GridArray, GridData, Array
        '''
        if callable(data):
            self._data[var] = data
            return

        if isinstance(data, (dm.GridArray, GridData)):
            if self._grid is None or not self._grid.initialized:
                self._grid = data.grid
            else:
                assert self._grid == data.grid

        if var in self._grid.vars:
            raise ValueError(
                'Variable `%s` is already a grid dimension!' % var
                )

        if isinstance(data, dm.GridData):
            assert len(data.data_vars) == 1
            data = data[0]

        if self.ndim == 0:
            print('adding default grid')
            # make a default grid:
            if data.ndim <= 3 and var not in ['x', 'y', 'z']:
                axes_names = ['x', 'y', 'z']
            else:
                axes_names = ['x%i' for i in range(data.ndim + 1)]
                axes_names.delete(var)
            axes = {}
            for d, n in zip(axes_names, data.shape):
                axes[d] = np.arange(n)
            self._grid = dm.Grid(**axes)

        if data.ndim < self.ndim and self.shape[-1] == 1:
            # add new axis
            data = data[..., np.newaxis]

        data = np.ma.asarray(data)

        if not data.shape[:self.ndim] == self.shape:
            raise ValueError(
                'Incompatible data of shape %s for grid of shape %s' %
                (data.shape, self.shape)
                )

        self._data[var] = data

    def get_array(self, var, flat=False):
        '''
        return bare array of data

        Parameters:
        -----------

        var : string
            variable to return
        flat : bool (optional)
            if true return flattened (1d) array
        '''
        if var in self._grid.vars:
            array = self._grid.point_meshgrid[self._grid.vars.index(var)]
        else:
            array = self._data[var]
        if flat:
            if array.ndim == self._grid.nax:
                return array.ravel()
            return array.reshape(self._grid.size, -1)

        return array

    def flat(self, var):
        '''return flatened-out array'''
        return self.get_array(var, flat=True)

    def __iter__(self):
        '''
        iterate over dimensions
        '''
        return iter([self[v] for v in self.data_vars])

    def items(self):
        return [
            (v, self[v]) for v in self.data_vars
            ]

    # --- Tranlsation methods ---

    def interp(self, *args, method=None, fill_value=np.nan, **kwargs):
        return translations.Interpolation(
            self, *args, method=method, fill_value=fill_value, **kwargs
            ).run()

    interp.__doc__ = translations.Interpolation.__init__.__doc__

    def histogram(self, *args, density=False, **kwargs):
        return translations.Histogram(self, *args, density=density,
                                      **kwargs).run()

    histogram.__doc__ = translations.Histogram.__init__.__doc__

    def binwise(self, *args, **kwargs):
        return dm.BinnedData(data=self, *args, **kwargs)

    def lookup(self, *args, **kwargs):
        return translations.Lookup(self, *args, **kwargs).run()

    lookup.__doc__ = translations.Lookup.__init__.__doc__

    def kde(self, *args, **kwargs):
        return translations.KDE(self, *args, **kwargs).run()

    kde.__doc__ = translations.KDE.__init__.__doc__

    def resample(self, *args, **kwargs):
        return translations.Resample(self, *args, **kwargs).run()

    resample.__doc__ = translations.Resample.__init__.__doc__

    def gaussian_filter(self, *args, **kwargs):
        return translations.gaussian_filter(self, *args, **kwargs).run()

    gaussian_filter.__doc__ = translations.gaussian_filter.__init__.__doc__

    def convolve(self, *args, **kwargs):
        return translations.convolve(self, *args, **kwargs).run()

    convolve.__doc__ = translations.convolve.__init__.__doc__

    def correlate(self, *args, **kwargs):
        return translations.correlate(self, *args, **kwargs).run()

    correlate.__doc__ = translations.correlate.__init__.__doc__

    def gaussian_laplace(self, *args, **kwargs):
        return translations.gaussian_laplace(self, *args, **kwargs).run()

    gaussian_laplace.__doc__ = translations.gaussian_laplace.__init__.__doc__

    def generic_filter(self, *args, **kwargs):
        return translations.generic_filter(self, *args, **kwargs).run()

    generic_filter.__doc__ = translations.generic_filter.__init__.__doc__

    def laplace(self, *args, **kwargs):
        return translations.laplace(self, *args, **kwargs).run()

    laplace.__doc__ = translations.laplace.__init__.__doc__

    def maximum_filter(self, *args, **kwargs):
        return translations.maximum_filter(self, *args, **kwargs).run()

    maximum_filter.__doc__ = translations.maximum_filter.__init__.__doc__

    def median_filter(self, *args, **kwargs):
        return translations.median_filter(self, *args, **kwargs).run()

    median_filter.__doc__ = translations.median_filter.__init__.__doc__

    def minimum_filter(self, *args, **kwargs):
        return translations.minimum_filter(self, *args, **kwargs).run()

    minimum_filter.__doc__ = translations.minimum_filter.__init__.__doc__

    def percentile_filter(self, *args, **kwargs):
        return translations.percentile_filter(self, *args, **kwargs).run()

    percentile_filter.__doc__ = translations.percentile_filter.__init__.__doc__

    def prewitt(self, *args, **kwargs):
        return translations.prewitt(self, *args, **kwargs).run()

    prewitt.__doc__ = translations.prewitt.__init__.__doc__

    def rank_filter(self, *args, **kwargs):
        return translations.rank_filter(self, *args, **kwargs).run()

    rank_filter.__doc__ = translations.rank_filter.__init__.__doc__

    def sobel(self, *args, **kwargs):
        return translations.sobel(self, *args, **kwargs).run()

    sobel.__doc__ = translations.sobel.__init__.__doc__

    def uniform_filter(self, *args, **kwargs):
        return translations.uniform_filter(self, *args, **kwargs).run()

    uniform_filter.__doc__ = translations.uniform_filter.__init__.__doc__

    # --- Plotting methods ---

    def plot(self, var=None, **kwargs):
        if var is None and len(self.data_vars) == 1:
            var = self.data_vars[0]
        if self.ndim == 1:
            return self.plot_step(var, **kwargs)
        elif self.ndim == 2:
            return self.plot_map(var, **kwargs)

    def plot_map(self, var=None, cbar=False, fig=None, ax=None, **kwargs):
        '''
        plot a variable as a map

        ax : pyplot axes object
        var : str
        '''
        if var is None:
            if len(self.data_vars) == 1:
                var = self.data_vars[0]
            else:
                raise ValueError('Need to specify variable to plot')
        if self._grid.nax == 2:
            return dm.plotting.plot_map(
                self[var], label=var, cbar=cbar, fig=fig, ax=ax, **kwargs
                )

        raise ValueError('Can only plot maps of 2d grids')

    def plot_contour(self, var=None, fig=None, ax=None, **kwargs):
        if var is None and len(self.data_vars) == 1:
            var = self.data_vars[0]
        return dm.plotting.plot_contour(self, var, fig=fig, ax=ax, **kwargs)

    def plot_step(self, var=None, fig=None, ax=None, **kwargs):
        if var is None and len(self.data_vars) == 1:
            var = self.data_vars[0]
        return dm.plotting.plot_step(
            self[var], label=var, fig=fig, ax=ax, **kwargs
            )

    plot_bands = dm.plotting.plot_bands

    def plot_errorband(self, var, errors, fig=None, ax=None, **kwargs):
        if var is None and len(self.data_vars) == 1:
            var = self.data_vars[0]
        return dm.plotting.plot_errorband(
            self, var, errors, fig=fig, ax=ax, **kwargs
            )



from __future__ import absolute_import
from collections.abc import Iterable
import copy
import numpy as np
import dama as dm
from dama import translations
from dama.utils.formatter import format_table
import dama.plotting

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


def wrap(func):
    '''wrapper function to translate named axis to int indices
    and to re-pack results into correct form'''
    def wrapped_func(*args, **kwargs):
        # find first instance of GridArray:
        first = None
        inputs = []
        grid = None
        for arg in args:
            if isinstance(arg, dm.GridArray):
                inputs.append(np.ma.asarray(arg))
                if first is None:
                    first = arg
                    grid = arg.grid
                else:
                    # make sure all grids are compatible
                    assert arg.grid == grid, 'Incompatible grids'
            else:
                inputs.append(arg)

        if first is None:
            raise ValueError()
        if 'axis' in kwargs:
            axis = kwargs.get('axis')
            if not isinstance(axis, tuple) and axis is not None:
                axis = (axis, )
            if axis is not None:
                new_axis = []
                for a in axis:
                    # translate them
                    if isinstance(a, str):
                        a = first.grid.vars.index(a)
                    if a < 0:
                        a += first.ndim
                    new_axis.append(a)
                if len(new_axis) == 1:
                    kwargs['axis'] = new_axis[0]
                else:
                    kwargs['axis'] = tuple(new_axis)
                axis = new_axis
        else:
            axis = None

        result = func(*inputs, **kwargs)
        if isinstance(result, (np.ma.masked_array, np.ndarray)):
            if result.ndim > 0:
                # new grid
                if axis is not None and any(
                    [a < first.grid.nax for a in axis]
                    ):
                    new_grid = copy.deepcopy(first.grid)
                    for a in sorted(axis)[::-1]:
                        # need to be careful, and start deleting from last element
                        if a < first.grid.nax:
                            del (new_grid.axes[a])
                else:
                    new_grid = first.grid

                new_obj = dm.GridArray(result, grid=new_grid)
                if new_obj.nax == 0:
                    return new_obj.data
                return new_obj
            if result.ndim == 0:
                return np.asscalar(result)
        return result

    return wrapped_func


class GridArray(np.ma.MaskedArray):
    '''Structure to hold a single GridArray

    Parameters
    ----------
    input_array : ndarray
    grid : dm.Grid (optional)
        if not specified, *args and **kwargs will be used to constrcut grid
        if those are also not specified, default grid is added
    '''
    def __new__(cls, input_array, *args, grid=None, **kwargs):
        # ToDo: sort out kwargs
        dtype = kwargs.pop('dtype', None)
        order = kwargs.pop('order', 'K')
        subok = kwargs.pop('subok', False)
        ndmin = kwargs.pop('ndmin', 0)

        super().__new__(
            cls,
            input_array,
            dtype=dtype,
            order=order,
            subok=subok,
            ndmin=ndmin
            )

        obj = np.ma.asarray(input_array).view(cls)
        if grid is not None:
            obj.grid = grid
        elif not len(args) == 0 or not len(kwargs) == 0:
            obj.grid = dm.Grid(*args, **kwargs)
        else:
            # make a default grid:
            if input_array.ndim <= 3:
                axes_names = ['x', 'y', 'z']
            else:
                axes_names = ['x%i' for i in range(input_array.ndim)]
            axes = {}
            for d, n in zip(axes_names, input_array.shape):
                axes[d] = np.arange(n)
            obj.grid = dm.Grid(**axes)
        return obj

    def __array_finalize__(self, obj, *args, **kwargs):
        super().__array_finalize__(obj)
        #print('finalize: args, kwargs', args, kwargs)
        if obj is None:
            print('obj none')
            return
        self.grid = getattr(obj, 'grid', None)
        return self

    def __reduce__(self):
        pickled_state = super().__reduce__()
        new_state = pickled_state[2] + (self.grid,)
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):
        self.grid = state[-1]  # Set the info attribute
        super().__setstate__(state[0:-1])

    def __repr__(self):
        return format_table(self, tablefmt='plain')

    def _repr_html_(self):
        '''for jupyter'''
        return format_table(self, tablefmt='html')

    def __str__(self):
        return format_table(self, tablefmt='plain')

    @property
    def nax(self):
        return self.grid.nax

    def __getitem__(self, item, *args):
        if isinstance(item, str):
            if item in self.grid.vars:
                data = self.get_array(item)
                new_data = dm.GridArray(data, grid=self.grid)
                return new_data
            else:
                raise IndexError('No variable %s in DataSet' % item)

        if isinstance(item, dm.GridArray):
            if item.dtype == np.bool:
                mask = np.logical_and(~self.mask, ~np.asarray(item))
                new_item = dm.GridArray(np.asarray(self), grid=self.grid)
                new_item.mask = mask
                return new_item
            raise NotImplementedError('get item %s' % item)
        if not isinstance(item, tuple):  # and not isinstance(item, slice):
            return self[(item, )]
        if isinstance(item, list):
            if all([isinstance(i, int) for i in item]):
                return self[(list, )]
            else:
                raise IndexError('Cannot process list of indices %s' % item)
        if isinstance(item, tuple):

            item = self.grid.convert_slice(item)

            new_grid = self.grid[item]
            if len(new_grid) == 0:
                # then we have a single element
                return np.ma.asarray(self)[item]
            return dm.GridArray(np.ma.asarray(self)[item], grid=new_grid)

    def __getattr__(self, item):
        try:
            return self[item]
        except Exception as e:
            raise AttributeError from e

    def get_array(self, var, flat=False):
        '''
        return bare array of data

        Parameters:
        -----------

        var : string
            variable to return of grid vars
        flat : bool (optional)
            if true return flattened (1d) array
        '''
        assert var in self.grid.vars
        array = self.grid.point_meshgrid[self.grid.vars.index(var)]
        if flat:
            if array.ndim == self.grid.nax:
                return array.ravel()
            return array.reshape(self.grid.size, -1)

        return array

    @property
    def array_shape(self):
        '''
        shape of array
        '''
        return self.shape

    def __setitem__(self, item, val):
        if isinstance(item, dm.GridArray):
            if item.dtype == np.bool:
                mask = np.logical_and(~self.mask, ~np.asarray(item))
                if np.isscalar(val):
                    self[item].data[mask] = val
                else:
                    self[item].data[mask] = val.data[mask]
                return
        if np.isscalar(self[item]):
            self.data[item] = val
            return
        if isinstance(item, list) and all([isinstance(i, int) for i in item]):
            self.data[item] = val
            return
        mask = ~self[item].mask
        if np.isscalar(val):
            self[item].data[mask] = val
        else:
            self[item].data[mask] = val.data[mask]

    @property
    def T(self):
        '''transpose'''
        if self.ndim > self.nax + 1:
            raise NotImplementedError()
        if self.nax == 1:
            return self
        if self.nax > 1:
            new_data = self.data.T
        if self.ndim == self.nax + 1:
            new_data = np.rollaxis(new_data, 0, self.ndim)
        return dm.GridArray(new_data, grid=self.grid.T)

    @wrap
    def __add__(self, other):
        return np.ma.add(self, other)

    @wrap
    def __radd__(self, other):
        return np.ma.add(other, self)

    @wrap
    def __sub__(self, other):
        return np.ma.subtract(self, other)

    @wrap
    def __rsub__(self, other):
        return np.ma.subtract(other, self)

    @wrap
    def __mul__(self, other):
        return np.ma.multiply(self, other)

    @wrap
    def __rmul__(self, other):
        return np.ma.multiply(other, self)

    @wrap
    def __truediv__(self, other):
        return np.ma.divide(self, other)

    @wrap
    def __rtruediv__(self, other):
        return np.ma.divide(other, self)

    @wrap
    def __pow__(self, other):
        return np.ma.power(self, other)

    @wrap
    def __rpow__(self, other):
        return np.ma.power(other, other)

    @wrap
    def __lt__(self, other):
        return np.ma.less(self, other)

    @wrap
    def __le__(self, other):
        return np.ma.less_equal(self, other)

    @wrap
    def __eq__(self, other):
        return np.ma.equal(self, other)

    @wrap
    def __ne__(self, other):
        return np.ma.not_equal(self, other)

    @wrap
    def __gt__(self, other):
        return np.ma.greater(self, other)

    @wrap
    def __ge__(self, other):
        return np.ma.greater_equal(self, other)

    @wrap
    def sum(self, **kwargs):
        return np.ma.sum(self, **kwargs)

    @wrap
    def mean(self, **kwargs):
        return np.ma.mean(self, **kwargs)

    @wrap
    def std(self, **kwargs):
        return np.ma.std(self, **kwargs)

    @wrap
    def average(self, **kwargs):
        return np.ma.average(self, **kwargs)

    @wrap
    def median(self, **kwargs):
        return np.ma.median(self, **kwargs)

    @wrap
    def min(self, **kwargs):
        return np.ma.min(self, **kwargs)

    @wrap
    def max(self, **kwargs):
        return np.ma.max(self, **kwargs)

    @wrap
    def cumsum(self, **kwargs):
        return np.ma.cumsum(self, **kwargs)

    @wrap
    def quantile(self, q, **kwargs):
        return np.swapaxes(np.quantile(self, q, **kwargs), 0, -1)

    @wrap
    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        '''callable for numpy ufuncs'''
        #print('ufunc')
        return np.ma.asarray(self).__array_ufunc__(
            ufunc, method, *inputs, **kwargs
            )

    def __array__(self):
        print('array')

    def __array_prepare__(self, result, context=None):
        print('prepare')
        return result

    def __array_wrap__(self, out_arr, context=None):
        print('In __array_wrap__:')
        #print('   self is %s' % repr(self))
        #print('   arr is %s' % repr(out_arr))
        obj = np.ma.asarray(out_arr).view(dm.GridArray)
        obj.grid = self.grid
        return obj

    def flat(self):
        '''return values as flattened array'''
        if self.ndim == self.nax:
            return np.ma.asarray(self).ravel()
        return np.ma.asarray(self).reshape(self.grid.size, -1)

    # --- Tranlsation methods ---

    def interp(self, *args, method=None, fill_value=np.nan, **kwargs):
        return translations.Interpolation(
            self, *args, method=method, fill_value=fill_value, **kwargs
            ).run()

    interp.__doc__ = translations.Interpolation.__init__.__doc__

    def histogram(self, *args, density=False, **kwargs):
        return translations.Histogram(self, *args, density=density,
                                      **kwargs).run()

    histogram.__doc__ = translations.Histogram.__init__.__doc__

    def binwise(self, *args, **kwargs):
        return dm.BinnedData(data=self, *args, **kwargs)

    def lookup(self, *args, **kwargs):
        return translations.Lookup(self, *args, **kwargs).run()

    lookup.__doc__ = translations.Lookup.__init__.__doc__

    def kde(
        self,
        *args,
        bw='silverman',
        kernel='gaussian',
        density=True,
        **kwargs
        ):
        return translations.KDE(
            self, *args, bw=bw, kernel=kernel, density=density, **kwargs
            ).run()

    kde.__doc__ = translations.KDE.__init__.__doc__

    def resample(self, *args, **kwargs):
        return translations.Resample(self, *args, **kwargs).run()

    resample.__doc__ = translations.Resample.__init__.__doc__

    def gaussian_filter(self, *args, **kwargs):
        return translations.gaussian_filter(self, *args, **kwargs).run()

    gaussian_filter.__doc__ = translations.gaussian_filter.__init__.__doc__

    def convolve(self, *args, **kwargs):
        return translations.convolve(self, *args, **kwargs).run()

    convolve.__doc__ = translations.convolve.__init__.__doc__

    def correlate(self, *args, **kwargs):
        return translations.correlate(self, *args, **kwargs).run()

    correlate.__doc__ = translations.correlate.__init__.__doc__

    def gaussian_laplace(self, *args, **kwargs):
        return translations.gaussian_laplace(self, *args, **kwargs).run()

    gaussian_laplace.__doc__ = translations.gaussian_laplace.__init__.__doc__

    def generic_filter(self, *args, **kwargs):
        return translations.generic_filter(self, *args, **kwargs).run()

    generic_filter.__doc__ = translations.generic_filter.__init__.__doc__

    def laplace(self, *args, **kwargs):
        return translations.laplace(self, *args, **kwargs).run()

    laplace.__doc__ = translations.laplace.__init__.__doc__

    def maximum_filter(self, *args, **kwargs):
        return translations.maximum_filter(self, *args, **kwargs).run()

    maximum_filter.__doc__ = translations.maximum_filter.__init__.__doc__

    def median_filter(self, *args, **kwargs):
        return translations.median_filter(self, *args, **kwargs).run()

    median_filter.__doc__ = translations.median_filter.__init__.__doc__

    def minimum_filter(self, *args, **kwargs):
        return translations.minimum_filter(self, *args, **kwargs).run()

    minimum_filter.__doc__ = translations.minimum_filter.__init__.__doc__

    def percentile_filter(self, *args, **kwargs):
        return translations.percentile_filter(self, *args, **kwargs).run()

    percentile_filter.__doc__ = translations.percentile_filter.__init__.__doc__

    def prewitt(self, *args, **kwargs):
        return translations.prewitt(self, *args, **kwargs).run()

    prewitt.__doc__ = translations.prewitt.__init__.__doc__

    def rank_filter(self, *args, **kwargs):
        return translations.rank_filter(self, *args, **kwargs).run()

    rank_filter.__doc__ = translations.rank_filter.__init__.__doc__

    def sobel(self, *args, **kwargs):
        return translations.sobel(self, *args, **kwargs).run()

    sobel.__doc__ = translations.sobel.__init__.__doc__

    def uniform_filter(self, *args, **kwargs):
        return translations.uniform_filter(self, *args, **kwargs).run()

    uniform_filter.__doc__ = translations.uniform_filter.__init__.__doc__
    # --- Plotting ---

    def plot(self, **kwargs):
        if self.nax == 1:
            return self.plot_step(**kwargs)
        elif self.nax == 2:
            return self.plot_map(**kwargs)
        else:
            raise NotImplementedError()

    def plot_map(self, label=None, cbar=False, fig=None, ax=None, **kwargs):
        '''
        plot array as a map

        ax : pyplot axes object
        '''
        if self.nax == 2:
            return dm.plotting.plot_map(
                self, label=label, cbar=cbar, fig=fig, ax=ax, **kwargs
                )

        raise ValueError('Can only plot maps of 2d grids')

    def plot_contour(self, fig=None, ax=None, **kwargs):
        return dm.plotting.plot_contour(self, fig=fig, ax=ax, **kwargs)

    def plot_step(self, label=None, fig=None, ax=None, **kwargs):
        return dm.plotting.plot_step(
            self, label=label, fig=fig, ax=ax, **kwargs
            )

    plot_bands = dm.plotting.plot_bands



from __future__ import absolute_import
'''Module providing a data translation methods'''
import numpy as np
from dama.translations import Translation

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class Resample(Translation):
    def __init__(self, source, *args, **kwargs):
        '''resample from binned data into other binned data
        ToDo: this is super inefficient
        '''
        super().__init__(
            source,
            *args,
            source_needs_grid=True,
            dest_needs_grid=True,
            **kwargs
            )

        assert self.dest.grid.vars == self.source.grid.vars, 'grid variables of source and destination must be identical'

    def setup(self):

        # we need a super sample of points, i.e. meshgrids of all combinations of source and dest
        # so first create for every dest.grid.var a vector of both, src and dest points
        lookup_sample = [
            np.concatenate(
                [self.dest.grid[var].points, self.source.grid[var].points]
                ) for var in self.wrt
            ]
        mesh = np.meshgrid(*lookup_sample)
        self.lookup_sample = [m.flatten() for m in mesh]
        self.indices = self.source.grid.compute_indices(self.lookup_sample)

    def eval(self, source_data):

        lookup_array = np.full(self.lookup_sample[0].shape[0], np.nan)
        for i in range(len(lookup_array)):
            # check we're inside grid:
            idx = self.indices[i]
            if idx >= 0:
                ind = np.unravel_index(idx, self.source.grid.shape)
                lookup_array[i] = source_data[idx]

        # now bin both these points into destination
        bins = self.dest.grid.squeezed_edges
        lu_hist, _ = np.histogramdd(
            sample=self.lookup_sample, bins=bins, weights=lookup_array
            )
        lu_counts, _ = np.histogramdd(sample=self.lookup_sample, bins=bins)
        lu_hist /= lu_counts

        return lu_hist



from __future__ import absolute_import
from numbers import Number
from collections import OrderedDict
from collections.abc import Iterable
import copy

import dama as dm

import numpy as np

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class Grid(object):
    '''
    Class to hold a number of axes
    '''

    __slots__ = ['axes',]

    def __init__(self, *args, **kwargs):
        '''
        Paramters:
        ----------
        args : Axis or Grid object, or list thereof

        an axis can also be given by kwargs
        kwargs : str,Number or str,array

        '''
        self.axes = []

        for d in args:
            self.add_axis(d)

        for k, v in kwargs.items():
            self[k] = v


    def __getattr__(self, item):
        try:
            return self[item]
        except Exception as e:
            raise AttributeError from e

    def __setattr__(self, item, value):
        if item in self.__slots__:
            object.__setattr__(self, item, value)
        else:
            self[item] = value

    def __getstate__(self):
        return self.axes

    def __setstate__(self, state):
        self.axes = state

    def __deepcopy__(self, memo=None):
        return dm.Grid(*copy.deepcopy(self.axes, memo=memo))


    def initialize(self, source=None):
        '''Method to initialize the grid if grid is not fully set up
        it derive information from source
        
        Parameters
        ----------
        source : dm.GridData, dm.PointData
        
        '''
        # ToDo: maybe do something smarter with default nbins?
        # check dest grid is set up, otherwise do so
        for var in self.vars:
            if not self[var].initialized:
                if self[var].nbins is None:
                    self[var].nbins = 10
                if source is None:
                    if isinstance(self[var].nbins, int):
                        self[var].points = np.arange(self[var].nbins)
                        continue
                # check if it might be from a grid
                if isinstance(source, (dm.GridData, dm.GridArray)):
                    if var in source.grid.vars:
                        if isinstance(self[var].nbins, float):
                            # this measn we want to multiply the old nbins
                            new_nbins = int(
                                source.grid[var].nbins * self[var].nbins
                                )
                        else:
                            new_nbins = self[var].nbins
                        if source.grid[var].has_edges:
                            if source.grid[var].log:
                                self[var].edges = np.logspace(
                                    np.log(source.grid[var].edges.min()),
                                    np.log(source.grid[var].edges.max()), new_nbins + 1
                                    )
                            else:
                                self[var].edges = np.linspace(
                                    source.grid[var].edges.min(),
                                    source.grid[var].edges.max(), new_nbins + 1
                                    )
                        if source.grid[var]._points is not None:
                            if source.grid[var].log:
                                self[var].points = np.logspace(
                                    np.log(source.grid[var].points.min()),
                                    np.log(source.grid[var].points.max()), new_nbins
                                    )
                            else:
                                self[var].points = np.linspace(
                                    source.grid[var].points.min(),
                                    source.grid[var].points.max(), new_nbins
                                    )
                        continue
                # in this case it's pointdata
                self[var].edges = np.linspace(
                    np.nanmin(source[var]), np.nanmax(source[var]),
                    self[var].nbins + 1
                    )

    def add_axis(self, axis):
        '''
        add aditional Axis

        Paramters:
        ----------
        axis : Axis or dict or basestring

        in case of a basestring, a new empty axisension gets added
        '''
        if isinstance(axis, dm.Axis):
            if axis.var in self.vars:
                self.axes[self.vars.index(axis.var)] = axis
            else:
                self.axes.append(axis)
        elif isinstance(axis, dict):
            axis = dm.Axis(**axis)
            self.add_axis(axis)
        elif isinstance(axis, str):
            new_axis = dm.Axis(var=axis)
            self.add_axis(new_axis)
        else:
            raise TypeError('Cannot add type %s' % type(axis))

    def __eq__(self, other):
        if not type(self) == type(other):
            return False
        equal = self.vars == other.vars
        return equal and all([self[var] == other[var] for var in self.vars])

    @property
    def T(self):
        '''transpose'''
        return dm.Grid(*list(self)[::-1])

    @property
    def initialized(self):
        '''
        wether the grid is set or not
        '''
        return self.nax > 0 and all([d.initialized for d in self])

    @property
    def regular(self):
        '''true is all axisensions are reguallarly spaced'''
        return all([d.regular for d in self])

    @property
    def consecutive(self):
        '''true is all edges are consecutive'''
        return all([d.edges.consecutive for d in self])

    @property
    def nax(self):
        '''
        number of grid axisensions
        '''
        return len(self.axes)

    @property
    def vars(self):
        '''
        grid axisension variables
        '''
        return [d.var for d in self]

    @property
    def edges(self):
        '''
        all edges
        '''
        return [axis.edges for axis in self]

    @property
    def squeezed_edges(self):
        '''
        all squeezed edges
        '''
        return [axis.squeezed_edges for axis in self]

    @property
    def points(self):
        '''
        all points
        '''
        return [axis.points for axis in self]

    @property
    def point_meshgrid(self):
        return np.meshgrid(*self.points, indexing='ij')

    @property
    def point_mgrid(self):
        return [d.T for d in self.point_meshgrid]

    @property
    def edge_meshgrid(self):
        return np.meshgrid(*self.squeezed_edges)

    @property
    def edge_mgrid(self):
        return [d.T for d in self.edge_meshgrid]

    @property
    def size(self):
        '''
        size = total number of bins / points
        '''
        return np.product([len(d) for d in self])

    def __len__(self):
        return self.nax

    def __str__(self):
        '''
        string representation
        '''
        strs = []
        for axis in self:
            strs.append('%s' % axis)
        return '\n'.join(strs)

    def __repr__(self):
        strs = []
        strs.append('Grid(')
        for axis in self:
            strs.append('%s,' % axis.__repr__())
        strs[-1] += ')'
        return '\n'.join(strs)

    def __iter__(self):
        '''
        iterate over axisensions
        '''
        return iter(self.axes)

    @property
    def shape(self):
        '''
        shape
        '''
        shape = []
        for axis in self:
            shape.append(len(axis))
        return tuple(shape)

    def __getitem__(self, item):
        '''
        item : int, str, slice, ierable
        '''
        if isinstance(item, str):
            # by name
            if not item in self.vars:
                raise IndexError('Variable %s not present'%item)
            idx = self.vars.index(item)
            return self.axes[idx]

        elif isinstance(item, Number):
            return self[(item, )]
        elif isinstance(item, slice):
            return self[(item, )]
        elif isinstance(item, list):
            if all([isinstance(i, str) for i in item]):
                new_obj = dm.Grid()
                for var in item:
                    new_obj.axes.append(self[var])
                return new_obj
            elif all([isinstance(i, (np.integer, int)) for i in item]):
                return self[(item, )]
            else:
                raise IndexError('Cannot process list of indices %s' % item)
        elif isinstance(item, tuple):
            if all([isinstance(i, str) for i in item]):
                return self[list(item)]

            item = self.convert_slice(item)

            new_obj = dm.Grid()
            for i in range(len(self)):
                if i < len(item):
                    assert item[i] is not Ellipsis
                    if isinstance(item[i], (np.integer, int)):
                        # we can skip this axisesnion, as it is one element
                        continue
                    new_obj.axes.append(self.axes[i][item[i]])
                else:
                    new_obj.axes.append(self.axes[i])
            return new_obj

        elif isinstance(item, Iterable):
            new_axes = []
            for it in item:
                new_axes.append(self[it])
            return dm.Grid(*new_axes)
        else:
            raise KeyError('Cannot get key from %s' % type(item))

    def convert_slice(self, idx):
        ''' convert indices/slices

        Parameters
        ----------
        idx : tuple
        '''
        new_idx = []
        for i in range(len(idx)):
            if i < self.nax:
                new_idx.append(self.axes[i].convert_slice(idx[i]))
            else:
                new_idx.append(idx[i])

        return tuple(new_idx)

    def __setitem__(self, item, val):
        self.add_axis({item: val})
        #raise AttributeError("to set a grid axisension, specify if it is `points` or `edges`, e.g.:\ngrid['%s'].edges = %s"%(item, val))

    def compute_indices(self, sample):
        '''
        calculate the bin indices for a a given sample as a raveled multi-index
        when values are outside any of the axes' binning, returns -1 as index
        '''
        if isinstance(sample, np.ndarray):
            assert sample.shape[0] == self.nax
        elif isinstance(sample, list):
            assert len(sample) == self.nax

        if not self.consecutive:
            raise NotImplementedError()

        # array holding raveld indices
        multi_index = [
            self.axes[i].compute_indices(sample[i]) for i in range(self.nax)
            ]
        # mask where any index is outside binning (= -1)
        mask = np.all([idx >= 0 for idx in multi_index], axis=0)
        if np.isscalar(mask):
            if not mask:
                return -1
            return np.ravel_multi_index(multi_index, self.shape)
        raveled_indices = np.full_like(multi_index[0], fill_value=-1)
        raveled_indices[mask] = np.ravel_multi_index(
            [idx[mask] for idx in multi_index], self.shape
            )
        return raveled_indices


def test():
    a = Grid(var='a', edges=np.linspace(0, 1, 2))
    print(a)
    print(a.vars)
    a.x = dm.Edges(np.linspace(0, 10, 11))
    a.y = np.logspace(-1, 1, 20)
    print(a['x'].points)
    print(a['x'].edges)
    print(a['x', 'y'])


if __name__ == '__main__':
    test()



from .core.grid import Grid
from .core.edges import Edges
from .core.axis import Axis
from .core.pointarray import PointArray
from .core.pointdata import PointData
from .core.gridarray import GridArray
from .core.griddata import GridData
from .core.binneddata import BinnedData
from . import plotting
from . import cm
from .utils.fileio import save, load



"""Collection of some statistics functions"""

import numpy as np

__license__ = '''Copyright 2020 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


def weighted_quantile(
    samples, weights, q, unify=False, interp='correct', method=7
    ):
    '''
    weighted quantile function
    
    Parameters:
    ----------
    samples : array
        points
    weights : array
        weights of points
    q : array
        quantiles to compute
    unify : bool (optional)
        add up weights of identical points, resulting in single points
    interp : string
        interpolation method, choice of ['new', 'const', 'linear']
        
    Notes:
    ------
    weights must be > 0 and < inf
    
    '''

    samples = np.asarray(samples)
    weights = np.asarray(weights)

    assert np.alltrue(weights > 0), 'Weights must be > 0'
    assert np.alltrue(weights < np.inf), 'Weights must be < inf'

    q = np.asarray(q)

    if unify:
        unique_samples, weight_idx = np.unique(samples, return_inverse=True)
        unique_weights = np.zeros(len(unique_samples))
        for i, j in enumerate(weight_idx):
            unique_weights[j] += weights[i]
        samples = unique_samples
        weights = unique_weights

    # sort
    sorted_idx = np.argsort(samples)
    samples = samples[sorted_idx]
    weights = weights[sorted_idx]

    # arithmetic mean
    widths = 0.5 * (weights[:-1] + weights[1:])
    # geometric mean
    #widths = np.sqrt(weights[:-1] * weights[1:])
    # harmonic mean
    #widths = 1/ (1/weights[:-1] + 1/weights[1:])
    widths /= np.sum(widths)

    edges = np.concatenate([[0.], np.cumsum(weights)])

    if method == 7:

        # compute weighted edges and midpoints
        edges -= 0.5 * weights[0]
        midpoints = 0.5 * (edges[:-1] + edges[1:])
        total_width = sum(weights) - weights[0] / 2 - weights[-1] / 2
        edges /= total_width
        midpoints = midpoints / total_width

    elif method == 5:

        edges /= np.sum(weights)
        midpoints = 0.5 * (edges[:-1] + edges[1:])

    if interp == 'expand':

        delta = (weights[1:] - weights[:-1])
        #weight_ratios = (weights[:-1] - weights[1:]) / (weights[:-1] + weights[1:])# lower / upper
        #weight_ratios = delta / (1 + np.abs(delta))
        weight_ratios = delta / np.maximum(weights[1:], weights[:-1])

        widths = np.diff(midpoints)

        new_x = np.zeros_like(weight_ratios)
        new_f = np.zeros_like(weight_ratios)

        mask = weight_ratios > 0
        # i.e. lower bigger
        new_x[mask] = midpoints[1:][mask] - weight_ratios[mask] * widths[
            mask]  #* 2/3
        new_f[mask] = samples[1:][mask]

        mask = ~mask
        # i.e. upper bigger
        new_x[mask] = midpoints[:-1][mask] - weight_ratios[mask] * widths[
            mask]  #* 2/3
        new_f[mask] = samples[:-1][mask]

        xp = np.empty(2 * len(midpoints) - 1)
        fp = np.empty_like(xp)

        xp[::2] = midpoints
        fp[::2] = samples
        xp[1::2] = new_x
        fp[1::2] = new_f

        return np.interp(q, xp, fp)

    if interp == 'weighted fraction':
        d = np.digitize(q, midpoints, right=True)

        # include left
        d[d == 0] = 1
        d -= 1

        # include right
        d[d == len(midpoints) - 1] = len(midpoints) - 2

        # x = partial distance between points
        x = (q - midpoints[d]) / (midpoints[d + 1] - midpoints[d])

        out = (
            (1 - x) * weights[d] * samples[d] +
            x * weights[d + 1] * samples[d + 1]
            ) / ((1 - x) * weights[d] + x * weights[d + 1])

        out[q > midpoints[-1]] = samples[-1]
        out[q < midpoints[0]] = samples[0]

        return out

    if interp == 'correct':
        d = np.digitize(q, midpoints, right=True)

        # include left
        d[d == 0] = 1
        d -= 1

        # include right
        d[d == len(midpoints) - 1] = len(midpoints) - 2

        linear = np.interp(q, midpoints, samples)

        # partial quantiles
        pq = (q - midpoints[d]) / (midpoints[d + 1] - midpoints[d])

        x0 = samples[d]
        x1 = samples[d + 1]
        w0 = weights[d]
        w1 = weights[d + 1]

        out = (
            -np.sqrt(
                (
                    (2 * x0 * w1) / ((x0 - x1)**2 * (w1 + w0)) -
                    (2 * w0 * x1) / ((x0 - x1)**2 * (w1 + w0))
                    )**2 - 4 * (
                        w0 / ((x0 - x1)**2 * (w1 + w0)) - w1 /
                        ((x0 - x1)**2 * (w1 + w0))
                        ) * (
                            pq - (w0 * x0**2) / ((x0 - x1)**2 * (w1 + w0)) -
                            (x0**2 * w1) / ((x0 - x1)**2 * (w1 + w0)) +
                            (2 * w0 * x0 * x1) / ((x0 - x1)**2 * (w1 + w0))
                            )
                ) - (2 * x0 * w1) / ((x0 - x1)**2 * (w1 + w0)) +
            (2 * w0 * x1) / ((x0 - x1)**2 * (w1 + w0))
            ) / (
                2 * (
                    w0 / ((x0 - x1)**2 * (w1 + w0)) - w1 /
                    ((x0 - x1)**2 * (w1 + w0))
                    )
                )

        #return out
        mask = w0 == w1
        out[mask] = linear[mask]

        out[q > midpoints[-1]] = samples[-1]
        out[q < midpoints[0]] = samples[0]

        return out

    if interp == 'linear':
        return np.interp(q, midpoints, samples)

    if interp == 'nearest':
        d = np.digitize(q, edges, right=True)
        # include left
        d[d == 0] = 1
        d -= 1
        d = np.clip(d, a_min=0, a_max=len(samples) - 1)
        return samples[d]

    if interp == 'lower':
        d = np.digitize(q, midpoints, right=False)
        d -= 1
        d[d < 0] = 0
        return samples[d]

    elif interp == 'higher':
        d = np.digitize(q, midpoints, right=True)
        d[d >= len(samples)] = len(samples) - 1
        return samples[d]

    elif interp == 'midpoint':
        sample_midpoints = (
            weights[:-1] * samples[:-1] + weights[1:] * samples[1:]
            ) / (weights[:-1] + weights[1:])

        d = np.digitize(q, midpoints, right=True)
        d[d == 0] = 1
        d -= 1
        d[d == len(midpoints) - 1] = len(midpoints) - 2

        out = sample_midpoints[d]
        # edge cases
        if q[0] == 0:
            out[0] = samples[0]
        if q[-1] == 1:
            out[-1] = samples[-1]
        return out

    elif interp == 'corners':
        sample_midpoints = (
            weights[:-1] * samples[:-1] + weights[1:] * samples[1:]
            ) / (weights[:-1] + weights[1:])

        xp = np.empty(2 * len(midpoints) - 1)
        fp = np.empty_like(xp)

        xp[::2] = midpoints
        fp[::2] = samples
        xp[1::2] = edges[1:-1]
        fp[1::2] = sample_midpoints

        return np.interp(q, xp, fp)



from __future__ import absolute_import
'''Module providing a data translation methods'''
import numpy as np
from scipy import ndimage

from dama.translations import Translation

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class gaussian_filter(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.gaussian_filter(source_data, *self.args, **self.kwargs)

gaussian_filter.__init__.__doc__ = ndimage.gaussian_filter.__doc__


class convolve(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.convolve(source_data, *self.args, **self.kwargs)

convolve.__init__.__doc__ = ndimage.convolve.__doc__



class correlate(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.correlate(source_data, *self.args, **self.kwargs)

correlate.__init__.__doc__ = ndimage.correlate.__doc__


class gaussian_laplace(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.gaussian_laplace(source_data, *self.args, **self.kwargs)

gaussian_laplace.__init__.__doc__ = ndimage.gaussian_laplace.__doc__



class generic_filter(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.generic_filter(source_data, *self.args, **self.kwargs)

generic_filter.__init__.__doc__ = ndimage.generic_filter.__doc__



class laplace(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.laplace(source_data, *self.args, **self.kwargs)

laplace.__init__.__doc__ = ndimage.laplace.__doc__



class maximum_filter(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.maximum_filter(source_data, *self.args, **self.kwargs)

maximum_filter.__init__.__doc__ = ndimage.maximum_filter.__doc__



class median_filter(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.median_filter(source_data, *self.args, **self.kwargs)

median_filter.__init__.__doc__ = ndimage.median_filter.__doc__



class minimum_filter(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.minimum_filter(source_data, *self.args, **self.kwargs)

minimum_filter.__init__.__doc__ = ndimage.minimum_filter.__doc__



class percentile_filter(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.percentile_filter(source_data, *self.args, **self.kwargs)

percentile_filter.__init__.__doc__ = ndimage.percentile_filter.__doc__



class prewitt(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.prewitt(source_data, *self.args, **self.kwargs)

prewitt.__init__.__doc__ = ndimage.prewitt.__doc__



class rank_filter(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.rank_filter(source_data, *self.args, **self.kwargs)

rank_filter.__init__.__doc__ = ndimage.rank_filter.__doc__



class sobel(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.sobel(source_data, *self.args, **self.kwargs)

sobel.__init__.__doc__ = ndimage.sobel.__doc__



class uniform_filter(Translation):
    def __init__(
        self, source, *args, **kwargs
        ):

        self.args = args
        self.kwargs = kwargs

        if not source.grid.regular:
            print("Warning: this method assumes regular grids!")

        super().__init__(source,
                         source,
                         source_needs_grid=True,
                         dest_needs_grid=True,
                         )

    def setup(self):
        self.prepare_source_sample()
        self.prepare_dest_sample(flat=False)

    def eval(self, source_data):
        return ndimage.uniform_filter(source_data, *self.args, **self.kwargs)

uniform_filter.__init__.__doc__ = ndimage.uniform_filter.__doc__





""" Module for table formatting """
from __future__ import absolute_import

from numbers import Number
import html
import numpy as np
import tabulate

import dama as dm

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''

N_MAX = 6
'''max rows to display in html formatting'''
PRECISION = 3
'''signfificant figures to dispay'''


def as_str(a):
    '''simple string formatter
    
    Parameters
    ----------
    a : Array, Number

    Retrurns
    --------
    string : str
    '''
    if isinstance(a, np.ma.core.MaskedConstant):
        return str(a)
    if isinstance(a, Number):
        return ('%.' + str(PRECISION) + 'g') % a
    return np.array2string(
        np.ma.asarray(a), precision=PRECISION, threshold=2, edgeitems=2
        )


def make_table_labels(axis, bold, brk):
    '''generate gird labels

    Parameters
    ----------
    axis : dm.Axis

    Returns
    -------
    labels : list of str
    '''
    if axis._points is None:
        if len(axis) <= N_MAX:
            return [bold % as_str(axis.edges[i]) for i in range(len(axis))]
        labels = [bold % as_str(axis.edges[i]) for i in range(N_MAX // 2)]
        labels += ['...']
        labels += [
            bold % as_str(axis.edges[i])
            for i in range(len(axis) - N_MAX // 2, len(axis))
            ]
        return labels

    # ToDo: a bit too much copy-paste going on here...
    elif not axis.has_edges or axis._edges._edges is None:
        if len(axis) <= N_MAX:
            return [bold % as_str(axis.points[i]) for i in range(len(axis))]
        labels = [bold % as_str(axis.points[i]) for i in range(N_MAX // 2)]
        labels += ['...']
        labels += [
            bold % as_str(axis.points[i])
            for i in range(len(axis) - N_MAX // 2, len(axis))
            ]
        return labels

    if len(axis) <= N_MAX:
        return [
            bold % (
                '[%s | %s | %s]' % (
                    as_str(axis.edges[i, 0]
                           ), as_str(axis.points[i]), as_str(axis.edges[i, 1])
                    )
                ) for i in range(len(axis))
            ]
    labels = [
        bold % (
            '[%s | %s | %s]' % (
                as_str(axis.edges[i, 0]), as_str(axis.points[i]
                                                 ), as_str(axis.edges[i, 1])
                )
            ) for i in range(N_MAX // 2)
        ]
    labels += ['...']
    labels += [
        bold % (
            '[%s | %s | %s]' % (
                as_str(axis.edges[i, 0]), as_str(axis.points[i]
                                                 ), as_str(axis.edges[i, 1])
                )
            ) for i in range(len(axis) - N_MAX // 2, len(axis))
        ]
    return labels


def make_table_row(name, array, bold, brk):
    '''forma a simgle table row'''
    if name is not None:
        row = [bold % name]
    else:
        row = []
    array = np.ma.asarray(array)
    #if array.ndim == 0:
    #    return [as_str(array)]
    if array.shape[0] <= N_MAX:
        row += [as_str(v) for v in array]
    else:
        row += [as_str(v) for v in array[:N_MAX // 2]]
        row += ['...']
        row += [as_str(v) for v in array[-N_MAX // 2:]]

    return row


def make_2d_table(data, bold, brk):

    x_labels = make_table_labels(data.grid.axes[0], bold, brk)
    y_labels = make_table_labels(data.grid.axes[1], bold, brk)

    table = []
    table.append(
        [bold % ('%s \\ %s' % (data.grid.vars[1], data.grid.vars[0]))] +
        x_labels
        )
    for i in range(len(y_labels)):
        table.append([y_labels[i]] + [0] * (len(x_labels)))

    n_data_cols, n_data_rows = data.shape[:2]

    if n_data_rows > N_MAX:
        for i in range(1, min(n_data_cols, 2 * (N_MAX // 2) + 1) + 1):
            table[N_MAX // 2 + 1][i] = '...'
        data_row_idices = list(range(N_MAX // 2)) + list(
            range(n_data_rows - N_MAX // 2, n_data_rows)
            )
        table_row_idices = list(range(N_MAX // 2)) + list(
            range(N_MAX // 2 + 1, 2 * (N_MAX // 2) + 1)
            )
    else:
        data_row_idices = list(range(n_data_rows))
        table_row_idices = list(range(n_data_rows))

    if n_data_cols > N_MAX:
        for i in range(1, min(n_data_rows, 2 * (N_MAX // 2) + 1) + 1):
            table[i][N_MAX // 2 + 1] = '...'
        data_col_idices = list(range(N_MAX // 2)) + list(
            range(n_data_cols - N_MAX // 2, n_data_cols)
            )
        table_col_idices = list(range(N_MAX // 2)) + list(
            range(N_MAX // 2 + 1, 2 * (N_MAX // 2) + 1)
            )
    else:
        data_col_idices = list(range(n_data_cols))
        table_col_idices = list(range(n_data_cols))

    for i, r_idx in zip(table_row_idices, data_row_idices):
        for j, c_idx in zip(table_col_idices, data_col_idices):
            table[i + 1][j + 1] = get_item(data, (c_idx, r_idx), bold, brk)

    return table


def get_item(data, idx, bold, brk):
    '''Get a string formatted item from a GridArray or gridData object at index idx'''
    if isinstance(data, dm.GridArray):
        return as_str(data[idx])
    elif isinstance(data, dm.GridData):
        # collect all items
        all_data = []
        for n, d in data.items():
            all_data.append('%s = %s' % (n, as_str(d[idx])))
        return brk.join(all_data)


def unescape_tabulate(table, tablefmt):
    """wrapper for tabulate to unescape html"""
    if tablefmt == 'html':
        return html.unescape(tabulate.tabulate(table, tablefmt=tablefmt))
    return tabulate.tabulate(table, tablefmt=tablefmt)

def format_table(data, tablefmt='html'):

    if tablefmt == 'html':
        bold = '<b>%s</b>'
        brk = '<br>'
    else:
        bold = '%s'
        brk = '\n'

    if isinstance(data, dm.PointData):
        table = [make_table_row(n, d, bold, brk) for n, d in data.items()]
        return unescape_tabulate(table, tablefmt=tablefmt)

    if isinstance(data, dm.PointArray):
        table = [make_table_row(None, data, bold, brk)]
        return unescape_tabulate(table, tablefmt=tablefmt)

    if isinstance(data, dm.GridArray):
        if data.nax == 2:
            table = make_2d_table(data, bold, brk)

        elif data.nax == 1:
            table = []
            table.append(
                [bold % data.grid.vars[0]] +
                make_table_labels(data.grid.axes[0], bold, brk)
                )
            table.append(make_table_row('', data, bold, brk))

        else:
            return None

        return unescape_tabulate(table, tablefmt=tablefmt)

    if isinstance(data, dm.GridData):
        if data.grid.nax == 2:
            table = make_2d_table(data, bold, brk)

        elif data.ndim == 1:
            table = []
            table.append(
                [bold % data.grid.vars[0]] +
                make_table_labels(data.grid.axes[0], bold, brk)
                )
            for d in data.items():
                table.append(make_table_row(*d, bold, brk))

        return unescape_tabulate(table, tablefmt=tablefmt)

    return data.__repr__()



from setuptools import setup, find_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()

setup(
    name='dama',
    version='0.4.7',
    packages=find_packages(),
    license='Apache 2.0',
    author='Philipp Eller',
    author_email='peller.phys@gmail.com',
    url='https://github.com/philippeller/dama',
    description='Look at data in different ways',
    long_description=long_description,
    long_description_content_type='text/markdown',

    python_requires='>=3.6',

    install_requires=[
        'scipy>=0.17',
        'matplotlib>=2.0',
        'KDEpy',
        'tabulate',
        'numpy_indexed',
        'numpy>=1.11',
    ],
)



from __future__ import absolute_import
'''Module providing a data base class for translation methods'''

import numpy as np

from dama.translations import Translation

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class Histogram(Translation):
    def __init__(self, source, *args, density=False, **kwargs):
        '''translation into historgram form

        Parameters:
        -----------
        density : bool (optional)
            if True return histograms as densities
        '''
        super().__init__(source, *args, dest_needs_grid=True, **kwargs)

        self.density = density

        if density:
            self.additional_runs = {'density': None}
        else:
            self.additional_runs = {'counts': None}

    def setup(self):
        self.prepare_source_sample(stacked=False)

    def eval(self, source_data):

        if source_data is None:
            output_array, _ = np.histogramdd(
                sample=self.source_sample,
                bins=self.dest.grid.squeezed_edges,
                density=self.density
                )

        else:
            source_data = source_data.flat()

            if source_data.ndim > 1:
                output_array = self.get_empty_output_array(
                    source_data.shape[1:]
                    )
                for idx in np.ndindex(*source_data.shape[1:]):
                    output_array[(Ellipsis, ) + idx], _ = np.histogramdd(
                        sample=self.source_sample,
                        bins=self.dest.grid.squeezed_edges,
                        weights=source_data[(Ellipsis, ) + idx],
                        density=self.density
                        )
            else:
                output_array, _ = np.histogramdd(
                    sample=self.source_sample,
                    bins=self.dest.grid.squeezed_edges,
                    weights=source_data,
                    density=self.density
                    )

        return output_array



from __future__ import absolute_import
import numpy as np
import dama as dm
from matplotlib import pyplot as plt
'''Module to provide plotting convenience functions
to be used by data source classes
'''

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''

# === Modernized ===


def plot_bands(source, var=None, fig=None, ax=None, labels=None, filled=True, lines=False, **kwargs):
    '''
    plot band between the variable's values (expect each bin to have a 1d array)

    Parameters:
    -----------
    var : str
        Variable name ot be plotted (if source type is GridArry or GridData with
        a single variable, then that one is used by default)
    
    fig, ax : matplotlib figure and axis (optional)

    labels : iterable
        lables to add for the bands

    filled : bool
        Draw filled areas between bands

    lines : bool
        Draw lines at bands
    '''

    #ToDo: fix invalid values

    assert isinstance(source, (dm.GridData, dm.GridArray))
    assert source.grid.nax == 1

    if isinstance(source, dm.GridData):
        if var is None and len(source.data_vars) == 1:
            var = source.data_vars[0]
        data = np.ma.asarray(source[var])

    else:
        data = np.ma.asarray(source)

    if fig is None:
        fig = plt.gcf()
    if ax is None:
        ax = plt.gca()

    cmap = kwargs.pop('cmap', 'Blues')
    cmap = plt.get_cmap(cmap)

    n_points = data.shape[1]
    n_bands = (n_points + 1) // 2

    if lines:
        linestyles = kwargs.pop('linestyles', ['-']*n_bands)
        linecolors = kwargs.pop('linecolors', ['k']*n_bands)

    colors = cmap(np.linspace(0, 1, n_bands + 1))[1:]
    colors = kwargs.pop('colors', colors)

    grid_axis = source.grid.axes[0]

    for i in range(n_bands):
        upper_idx = n_points - i - 1
        if labels is not None and i < len(labels):
            label = labels[i]
        else:
            label = None

        if grid_axis.has_points:
            if not upper_idx == i:
                if filled:
                    ax.fill_between(
                        grid_axis.points,
                        data[:, i],
                        data[:, upper_idx],
                        color=colors[i],
                        label=label,
                        **kwargs
                        )
                if lines:
                    ax.plot(
                        grid_axis.points,
                        data[:, i],
                        color=linecolors[i],
                        linestyle=linestyles[i],
                        label=label,
                        **kwargs
                        )
                    ax.plot(
                        grid_axis.points,
                        data[:, upper_idx],
                        color=linecolors[i],
                        linestyle=linestyles[i],
                        **kwargs
                        )
            else:
                if filled:
                    ax.plot(
                        grid_axis.points,
                        data[:, i],
                        color=colors[i],
                        label=label,
                        **kwargs
                        )
                if lines:
                    ax.plot(
                        grid_axis.points,
                        data[:, i],
                        color=linecolors[i],
                        linestyle=linestyles[i],
                        label=label,
                        **kwargs
                        )

        else:
            if not upper_idx == i:
                if filled:
                    ax.bar(
                        grid_axis.edges.edges[:, 0],
                        data[:, upper_idx] - data[:, i],
                        bottom=np.nan_to_num(data[:, i]),
                        width=grid_axis.edges.width,
                        color=colors[i],
                        align='edge',
                        label=label,
                        **kwargs
                        )
                if lines:
                    band_data = np.ma.asarray(data[:, i])
                    band_data = np.ma.append(band_data, band_data[-1])
                    ax.step(
                        grid_axis.squeezed_edges,
                        band_data,
                        where='post',
                        label=label,
                        color=linecolors[i],
                        linestyle=linestyles[i],
                        **kwargs
                        )
                    band_data = np.ma.asarray(data[:, upper_idx])
                    band_data = np.ma.append(band_data, band_data[-1])
                    ax.step(
                        grid_axis.squeezed_edges,
                        band_data,
                        where='post',
                        color=linecolors[i],
                        linestyle=linestyles[i],
                        **kwargs
                        )
            else:
                band_data = np.ma.asarray(data[:, i])
                band_data = np.ma.append(band_data, band_data[-1])
                if filled:
                    ax.step(
                        grid_axis.squeezed_edges,
                        band_data,
                        where='post',
                        label=label,
                        color=colors[i],
                        **kwargs
                        )
                if lines:
                    ax.step(
                        grid_axis.squeezed_edges,
                        band_data,
                        where='post',
                        label=label,
                        color=linecolors[i],
                        linestyle=linestyles[i],
                        **kwargs
                        )

    ax.set_xlabel(source.grid.vars[0])
    if source.grid.axes[0].log:
        ax.set_xscale('log')
    ax.set_ylabel(var)

    if grid_axis.has_points:
        ax.set_xlim(grid_axis.points.min(), grid_axis.points.max())
    else:
        ax.set_xlim(grid_axis.edges.min(), grid_axis.edges.max())


def plot_map(source, var=None, cbar=False, fig=None, ax=None, **kwargs):
    '''
    plot a 2d color map

    Parameters:
    -----------

    var : str (optional)
        Variable name ot be plotted (if source type is GridArry or GridData with
        a single variable, then that one is used by default)
    cbar : bool (optional)
        Add colorbar to axis
    fig, ax : matplotlib figure and axis (optional)
    '''
    assert isinstance(source, (dm.GridData, dm.GridArray))
    assert source.grid.nax == 2

    if isinstance(source, dm.GridData):
        if var is None and len(source.data_vars) == 1:
            var = source.data_vars[0]
        data = source[var]

    else:
        data = source

    if fig is None:
        fig = plt.gcf()
    if ax is None:
        ax = plt.gca()

    data = np.ma.asarray(data)

    if data.ndim == source.grid.nax + 1 and data.shape[-1] == 3:
        # plot as image
        pc = ax.imshow(
            data.swapaxes(0, 1)[::-1, :, :],
            extent=(
                source.grid.edges[0].min(), source.grid.edges[0].max(),
                source.grid.edges[1].min(), source.grid.edges[1].max()
                ),
            **kwargs
            )
    else:
        X, Y = source.grid.edge_meshgrid
        pc = ax.pcolormesh(
            X, Y, data.T, linewidth=0, rasterized=True, **kwargs
            )
        if cbar:
            fig.colorbar(pc, ax=ax, label=kwargs.pop('label', var))

    ax.set_xlabel(source.grid.vars[0])
    if source.grid.axes[0].log:
        ax.set_xscale('log')
    ax.set_ylabel(source.grid.vars[1])
    if source.grid.axes[1].log:
        ax.set_yscale('log')
    ax.set_xlim(source.grid.edges[0].min(), source.grid.edges[0].max())
    ax.set_ylim(source.grid.edges[1].min(), source.grid.edges[1].max())
    return pc


def plot_step(source, var=None, label=None, fig=None, ax=None, **kwargs):
    '''
    plot a step function, i.e. histogram
    var : str
        Variable name ot be plotted (if source type is GridArry or GridData with
        a single variable, then that one is used by default)
    label : str
    fig, ax : matplotlib figure and axis (optional)
    '''
    assert isinstance(source, (dm.GridData, dm.GridArray))
    assert source.grid.nax == 1

    if isinstance(source, dm.GridData):
        if var is None and len(source.data_vars) == 1:
            var = source.data_vars[0]
        data = np.array(source[var])

    else:
        data = np.array(source)

    if fig is None:
        fig = plt.gcf()
    if ax is None:
        ax = plt.gca()

    data = np.ma.asarray(data)
    data = np.ma.append(data, data[-1])

    s = ax.step(
        source.grid.squeezed_edges[0],
        data,
        where='post',
        label=label,
        **kwargs
        )
    ax.set_xlabel(source.grid.vars[0])
    if source.grid.axes[0].log:
        ax.set_xscale('log')
    ax.set_ylabel(var)
    return s


def plot1d_all(source, *args, **kwargs):
    fig = kwargs.pop('fig', plt.gcf())
    ax = kwargs.pop('ax', plt.gca())

    if isinstance(source, dm.PointArray):
        return ax.plot(source)

    for var in source.vars:
        ax.plot(source[var], label=var)


# --- to be fixed ---


def plot1d(source, x, *args, **kwargs):
    '''1d plot'''
    fig = kwargs.pop('fig', plt.gcf())
    ax = kwargs.pop('ax', plt.gca())
    p = ax.plot(source[x], *args, **kwargs)
    ax.set_ylabel(x)
    return p


def plot(source, *args, labels=None, **kwargs):
    '''2d plot
    
    Parameters:
    -----------

    args[0] : str or Iterable (optional)
        data variables to plot 
    args[1] : string or Iterable (optional)
        data variables to plot resulting in 2d plots
    labels : string or Iterable (optional)
    
    '''

    x = None
    y = None
    if len(args) > 0:
        if isinstance(args[0], str) and args[0] not in source.vars:
            x = None
            y = None
        else:
            x = args[0]
            args = args[1:]

        if len(args) > 0:
            if isinstance(args[0], str) and args[0] not in source.vars:
                y = None
            else:
                y = args[0]
                args = args[1:]

    fig = kwargs.pop('fig', plt.gcf())
    ax = kwargs.pop('ax', plt.gca())

    if x is None and y is None:
        if isinstance(source, dm.PointArray):
            return ax.plot(source, *args, label=labels, **kwargs)

        for i, var in enumerate(source.vars):
            if labels is None:
                label = var
            else:
                label = labels[i]
            ax.plot(source[var], *args, label=label, **kwargs)

    elif y is None:
        if isinstance(x, str):
            ax.plot(source[x], *args, label=labels, **kwargs)
            ax.set_ylabel(x)
        else:
            for i, x_var in enumerate(x):
                if labels is not None:
                    label = labels[i]
                else:
                    label = x_var
                ax.plot(source[x_var], *args, label=label, **kwargs)

    elif isinstance(x, str):
        if isinstance(y, str):
            p = ax.plot(source[x], source[y], *args, label=labels, **kwargs)
            ax.set_xlabel(x)
            ax.set_ylabel(y)
            return p
        else:
            for i, y_var in enumerate(y):
                if labels is not None:
                    label = labels[i]
                else:
                    label = None
                ax.plot(source[x], source[y_var], *args, label=label, **kwargs)
            ax.set_xlabel(x)
    else:
        if isinstance(y, str):
            for i, x_var in enumerate(x):
                if labels is not None:
                    label = labels[i]
                else:
                    label = None
                ax.plot(source[x_var], source[y], *args, label=label, **kwargs)
            ax.set_ylabel(y)

        else:

            assert len(x) == len(
                y
                ), 'Need same length of x and y variables list'

            for i, (x_var, y_var) in enumerate(zip(x, y)):
                if labels is not None:
                    label = labels[i]
                else:
                    label = None
                ax.plot(
                    source[x_var], source[y_var], *args, label=label, **kwargs
                    )


def plot_points_2d(
    source, x, y, s=None, c=None, cbar=False, fig=None, ax=None, **kwargs
    ):
    '''2d scatter plot'''
    if fig is None:
        fig = plt.gcf()
    if ax is None:
        ax = plt.gca()
    if c is not None:
        c_label = c
        c = source[c]
    else:
        assert not cbar
    if s is not None:
        if isinstance(s, str):
            s = source[s]
    sc = ax.scatter(
        np.array(source[x]),
        np.array(source[y]),
        s=np.array(s),
        c=np.array(c),
        **kwargs
        )
    if cbar:
        fig.colorbar(sc, ax=ax, label=c_label)
    ax.set_xlabel(x)
    ax.set_ylabel(y)
    return sc


def plot_contour(source, var=None, fig=None, ax=None, **kwargs):
    '''
    contours from gird data
    '''
    assert isinstance(source, (dm.GridData, dm.GridArray))
    assert source.grid.nax == 2

    if isinstance(source, dm.GridData):
        if var is None and len(source.data_vars) == 1:
            var = source.data_vars[0]
        data = np.ma.asarray(source[var])

    else:
        data = np.ma.asarray(source)



    if fig is None:
        fig = plt.gcf()
    if ax is None:
        ax = plt.gca()
    X, Y = source.grid.point_meshgrid

    labels = kwargs.pop('labels', None)
    inline = kwargs.pop('inline', True)

    cs = ax.contour(X, Y, data, **kwargs)

    if labels is not None:
        fmt = {}
        for l, s in zip(cs.levels, labels):
            fmt[l] = s

        ax.clabel(cs, cs.levels, inline=inline, fmt=fmt)

    if source.grid.axes[0].log:
        ax.set_xscale('log')
    if source.grid.axes[1].log:
        ax.set_yscale('log')

    return cs


def plot_errorband(source, var, errors, fig=None, ax=None, **kwargs):
    '''
    plot a step histogram with errorbars around it as bands
    '''
    if fig is None:
        fig = plt.gcf()
    if ax is None:
        ax = plt.gca()
    if isinstance(errors, (tuple, list)):
        lower_error = source[errors[0]]
        upper_error = source[errors[1]]
    elif isinstance(errors, str):
        lower_error = source[errors]
        upper_error = lower_error
    else:
        raise TypeError(
            'errors must be tuple of variable names or a single variable name'
            )
    assert source.grid.nax == 1

    ax.bar(
        source.grid[0].points,
        lower_error + upper_error,
        bottom=source[var] - lower_error,
        width=source.grid[0].edges.width,
        **kwargs
        )
    ax.set_xlabel(source.grid[0].var)

    if source.grid.axes[0].log:
        ax.set_xscale('log')



from __future__ import absolute_import
'''Module providing a data translation methods'''
import numpy as np
import dama as dm
from dama.translations import Translation

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class Lookup(Translation):
    def __init__(self, source, *args, **kwargs):
        '''lookup the bin content at given points

        Parameters:
        -----------

        ToDo: use ndimage.map_coordinates for regular grids
        '''
        super().__init__(source, *args, source_needs_grid=True, **kwargs)

    def setup(self):
        self.prepare_dest_sample()
        self.indices = self.source.grid.compute_indices(self.dest_sample)

    def eval(self, source_data):
        output_array = self.get_empty_output_array(
            source_data.shape[source_data.nax:], flat=True
            )

        for i in range(len(output_array)):
            # check we're inside grid:
            idx = self.indices[i]
            if idx >= 0:
                ind = np.unravel_index(idx, self.source.grid.shape)
                output_array[i] = source_data[ind]
        if output_array.size > self.dest.size:
            return output_array.reshape((self.dest.array_shape) + (-1, ))
        return output_array.reshape(self.dest.array_shape)



from __future__ import absolute_import
from numbers import Number
from collections.abc import Iterable

import dama as dm

import numpy as np

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class Edges(object):
    '''Holding binning edges'''
    def __init__(self, *args, delta=None, points=None, log=None, **kwargs):
        '''
        Paramters
        ---------
        args : either (nbins,) or (min, max, nbins) or (array,)
        delta : float (optional):
            specify the the delta between two subsequent edges
        points : array (optional)
            create edges from points
        log : bool (optional)
            spacing is logarthmic

        '''
        #print(args, kwargs)
        self._edges = None
        self._log = log
        if points is not None:
            self._add_edges(self.edges_from_points(points))
        elif len(args) == 1 and isinstance(args[0], dm.Edges):
            self._add_edges(args[0].edges)
        elif len(args) == 1 and isinstance(args[0], np.ndarray):
            self._add_edges(args[0])
        elif len(args) == 1 and isinstance(args[0], list):
            self._add_edges(np.array(args[0]))
        elif len(args) == 3:
            if self.log:
                self._add_edges(np.logspace(np.log(args[0]), np.log(args[1]), args[2]))
            else:
                self._add_edges(np.linspace(*args))
        elif len(args) == 0 and delta is None and len(kwargs) == 0:
            pass
        elif len(args) == 1 and args[0] is None:
            pass
        else:
            raise NotImplementedError()

    @property
    def log(self):
        if self._log is not None:
            return self._log
        if self._edges is None:
            return False
        if np.all(self._edges > 0):
            d = np.log(self._edges[:, 1]) - np.log(self._edges[:, 0])
            return np.allclose(d[0], d)
        return False

    @log.setter
    def log(self, log):
        self._log = log


    def __array__(self):
        if len(self) == 1:
            return self._edges[0]
        return self._edges

    def __repr__(self):
        return 'edges: ' + self._edges.__repr__()

    def __str__(self):
        return self._edges.__str__()

    def edges_from_points(self, points):
        '''
        create edges around points
        '''
        if len(points) == 1:
            # in this case we cannot do a delta, just do some sensible default behaviour
            if self.log:
                return np.array([points[0]/2, points[0]*2])
            return np.array([[points[0] - 0.5, points[0] + 0.5]])
        if self.log:
            lp = np.log(points)
            diff = np.diff(lp) / 2.
            return np.exp(np.concatenate(
                [
                    [lp[0] - diff[0]], lp[:-1] + diff,
                    [lp[-1] + diff[-1]]
                    ]
                ))

        diff = np.diff(points) / 2.
        return np.concatenate(
            [
                [points[0] - diff[0]], points[:-1] + diff,
                [points[-1] + diff[-1]]
                ]
            )

    def min(self):
        return np.min(self.edges)

    def max(self):
        return np.max(self.edges)

    @property
    def points(self):
        '''
        create points from centers between edges
        '''
        if self.log:
            points = np.sqrt(self._edges[:,0] * self._edges[:,1])
        else:
            points = np.average(self._edges, axis=1)
        if isinstance(points, Number):
            return np.array(points)
        return points

    def _add_edges(self, edges):
        if edges.ndim == 2:
            self._edges = edges
        elif edges.ndim == 1:
            assert len(edges) > 1, 'Edges must be at least length 2'
            self._edges = np.empty((len(edges) - 1, 2))
            self._edges[:, 0] = edges[:-1]
            self._edges[:, 1] = edges[1:]
        else:
            raise ValueError()

    @property
    def consecutive(self):
        '''True if edges consecutive, i.e. no gaps'''
        return np.all(self._edges[1:, 0] == self._edges[:-1, 1])

    @property
    def regular(self):
        '''True if spacing of edges is regular'''
        if self.log:
            d = np.log(self._edges[:, 1]) - np.log(self._edges[:, 0])
        else:
            d = self._edges[:, 1] - self._edges[:, 0]
        return np.allclose(d[0], d)

    def __len__(self):
        if self._edges is None:
            return 0
        return self._edges.shape[0]

    @property
    def edges(self):
        '''return edges'''
        return self._edges

    @edges.setter
    def edges(self, val):
        self._add_edges(val)

    @property
    def width(self):
        '''return the bin width'''
        return np.abs(self._edges[:, 1] - self._edges[:, 0])

    @property
    def squeezed_edges(self):
        '''return edges in squeezed form
        that is a 1d array as used in historgam functions and the like
        only available for consecutive edges'''
        if self.consecutive:
            squeezed_edges = np.empty(len(self) + 1)
            squeezed_edges[:-1] = self._edges[:, 0]
            squeezed_edges[-1] = self._edges[-1, 1]
            return squeezed_edges
        else:
            raise ValueError(
                'Can only provide squeezed edges for consecutive edges'
                )

    def __getitem__(self, idx):
        if idx is Ellipsis:
            return self
        new_edges = self._edges[idx]
        if np.isscalar(new_edges):
            return new_edges
        return dm.Edges(new_edges)

    def __eq__(self, other):
        return np.all(np.equal(self._edges, other._edges))


def BayesianEdges(sample):
    """Bayesian Blocks Implementation

    By Jake Vanderplas.  License: BSD
    Based on algorithm outlined in http://adsabs.harvard.edu/abs/2012arXiv1207.5578S

    Edit: Small bugfix (P. Eller)

    Parameters
    ----------
    sample : ndarray, length N
        data to be histogrammed
        
    weights : ndarray, length N (optional)
        weights of data

    Returns
    -------
    bins : ndarray
        array containing the (N+1) bin edges

    Notes
    -----
    This is an incomplete implementation: it may fail for some
    datasets.  Alternate fitness functions and prior forms can
    be found in the paper listed above.
    """
    # copy and sort the array
    sample = np.sort(sample)
    N = sample.size

    # create length-(N + 1) array of cell edges
    edges = np.concatenate([sample[:1],
                            0.5 * (sample[1:] + sample[:-1]),
                            sample[-1:]])
    block_length = sample[-1] - edges

    
    #print(np.diff(edges))
    #print(block_length)
    
    # arrays needed for the iteration
    nn_vec = np.ones(N)
    best = np.zeros(N, dtype=float)
    last = np.zeros(N, dtype=int)

    
    #-----------------------------------------------------------------
    # Start with first data cell; add one cell at each iteration
    #-----------------------------------------------------------------
    for K in range(N):
        # Compute the width and count of the final bin for all possible
        # locations of the K^th changepoint
        width = block_length[:K + 1] - block_length[K + 1]
        count_vec = np.cumsum(nn_vec[:K + 1][::-1])[::-1]
        # evaluate fitness function for these possibilities
        
        fit_vec = count_vec * (np.log(count_vec) - np.log(width))
        fit_vec -= 4  # 4 comes from the prior on the number of changepoints
        fit_vec[1:] += best[:K]

        # find the max of the fitness: this is the K^th changepoint
        i_max = np.argmax(fit_vec)
        last[K] = i_max
        best[K] = fit_vec[i_max]
        
    #-----------------------------------------------------------------
    # Recover changepoints by iteratively peeling off the last block
    #-----------------------------------------------------------------
    change_points =  np.zeros(N+1, dtype=int)
    i_cp = N+1
    ind = N
    while True:
        i_cp -= 1
        change_points[i_cp] = ind
        if ind == 0:
            break
        ind = last[ind - 1]
        
    change_points = change_points[i_cp:]

    return dm.Edges(edges[change_points])

def QuantileEdges(sample, n=10):
    """
    Return binning edges each containing the same quantile of the sample
    """
    qs = np.linspace(0, 1, n+1)
    edges = np.quantile(sample, qs)
    return dm.Edges(edges)

def test_edges():
    edges = Edges(np.arange(10))
    assert np.allclose(edges.points, np.linspace(0.5, 8.5, 9))



from __future__ import absolute_import
'''Module providing a data base class for translation methods'''
import numpy as np
import dama as dm

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class Translation():
    '''Base class for translation methods'''
    def __init__(
        self,
        source,
        *args,
        source_needs_grid=False,
        dest_needs_grid=False,
        **kwargs
        ):

        self.source = source
        self.source_has_grid = isinstance(
            self.source, (dm.GridData, dm.GridArray, dm.Grid)
            )
        if source_needs_grid:
            assert self.source_has_grid, 'Source must provide grid'
        self.dest = self.generate_destination(*args, **kwargs)
        self.dest_has_grid = isinstance(
            self.dest, (dm.GridData, dm.GridArray, dm.Grid)
            )
        if dest_needs_grid:
            assert self.dest_has_grid, 'Destination must provide grid'

        if self.dest_has_grid:
            self.wrt = self.dest.grid.vars
        elif self.source_has_grid:
            self.wrt = self.source.grid.vars
        else:
            self.wrt = self.dest.vars

        # checks
        if not isinstance(self.source, dm.GridArray):
            if not set(self.wrt) <= set(source.vars):
                raise TypeError(
                    'the following variables are missing in the source: %s' %
                    ', '.
                    join(set(self.vars) - (set(self.vars) & set(source.vars)))
                    )

        self.source_sample = None
        self.dest_sample = None

        # if there are any special, additional runs to be performed,
        # for example for histograms and KDE without source_data
        self.additional_runs = {}

    def generate_destination(self, *args, **kwargs):
        '''Correctly set up a destination data format
        depending on the supplied input
        
        Parameters
        ----------
        args, kwargs

        '''
        if len(args) == 1 and len(kwargs) == 0:
            dest = args[0]
            if isinstance(dest, dm.GridData):
                grid = dest.grid
                grid.initialize(self.source)
                return dm.GridData(grid)
            if isinstance(dest, dm.Grid):
                grid = dest
                grid.initialize(self.source)
                return dm.GridData(grid)
            if isinstance(dest, dm.GridArray):
                grid = dest.grid
                grid.initialize(self.source)
                return dm.GridArray(np.empty(grid.shape), grid=grid)
            if isinstance(dest, dm.PointData):
                # check which vars we need:
                if self.source_has_grid:
                    return dest[self.source.grid.vars]
                else:
                    return dest

        # check if source has a grid and if any args are in there
        if isinstance(self.source, (dm.GridData, dm.GridArray)):
            dims = []
            for arg in args:
                # in this case the source may have a grid, get those edges
                if isinstance(arg, str):
                    if arg in self.source.grid.vars:
                        dims.append(self.source.grid[arg])
                        continue
                dims.append(arg)
            args = dims

        # instantiate
        grid = dm.Grid(*args, **kwargs)
        grid.initialize(self.source)

        if isinstance(self.source, dm.GridArray):
            return dm.GridArray(np.empty(grid.shape), grid=grid)

        return dm.GridData(grid)

    def prepare_source_sample(self, flat=True, stacked=True, transposed=False):
        if transposed: assert stacked
        self.source_sample = [
            self.source.get_array(var, flat=flat) for var in self.wrt
            ]
        if stacked:
            self.source_sample = np.stack(self.source_sample)
        if transposed:
            self.source_sample = self.source_sample.T

    def prepare_dest_sample(self, flat=True, stacked=True, transposed=False):
        if transposed: assert stacked
        self.dest_sample = [
            self.dest.get_array(var, flat=flat) for var in self.wrt
            ]
        if stacked:
            self.dest_sample = np.stack(self.dest_sample)
        if transposed:
            self.dest_sample = self.dest_sample.T

    def setup(self):
        pass

    def run(self):

        self.setup()

        if isinstance(self.dest, dm.GridArray):
            # in this case it is a single array, no vars
            if isinstance(self.source, (dm.GridData, dm.PointData)):
                assert len(self.source.vars) - len(self.wrt) == 1
                for var in self.source.vars:
                    if var in self.wrt:
                        continue
                    source_data = self.source[var]
            else:
                source_data = self.source

            result = self.eval(source_data)
            self.dest = dm.GridArray(result, grid=self.dest.grid)
            return self.dest

        if isinstance(self.source, dm.GridArray):
            source_data = self.source
            result = self.eval(source_data)
            self.dest['result'] = result
            return self.dest

        def make_closure(data):
            def f():
                return self.eval(data)
            return f

        for var in self.source.vars:
            if var in self.wrt:
                continue
            self.dest[var] = make_closure(self.source[var])

        for var, data in self.additional_runs.items():
            self.dest[var] = make_closure(data)


        return self.dest

    def eval(self, data):
        raise NotImplementedError('Translation method must implement this')

    def get_empty_output_array(
        self, element_shape=tuple(), fill_value=np.nan, flat=False
        ):
        '''make empty array in shape of destinaion
    
        element_shape : tuple
            additional dimensions of the output
        fill_value : value
            fill value
        flat : bool
            if True, make flat in array dimensions
        '''
        array_shape = self.dest.array_shape
        if flat:
            array_shape = tuple([np.product(array_shape)])
        array_shape += element_shape

        if fill_value is None:
            return np.empty(array_shape)
        return np.full(array_shape, fill_value)



from __future__ import absolute_import
from numbers import Number
from collections import OrderedDict
from collections.abc import Iterable

import dama as dm

import numpy as np

__license__ = '''Copyright 2019 Philipp Eller

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.'''


class Axis(object):
    '''
    Class to hold a single Axis of a Grid
    which can have points and/or edges
    '''
    def __init__(
        self, var=None, edges=None, points=None, nbins=None, log=None, label=None, **kwargs
        ):

        if len(kwargs) == 1:
            assert var is None and edges is None and points is None
            var, val = list(kwargs.items())[0]
            if isinstance(val, (list, np.ndarray)):
                points = np.asanyarray(val)
            elif isinstance(val, dm.Edges):
                edges = val
            elif isinstance(val, Number):
                nbins = val
            else:
                raise ValueError()

        self.var = var
        self.label = label
        self._edges = edges
        self._points = points
        self._nbins = nbins
        self._log = log

    @property
    def has_points(self):
        '''True if points are set'''
        return self._points is not None

    @property
    def log(self):
        if self._log is not None:
            return self._log
        if self.has_edges:
            return self._edges.log
        if self.has_points and np.all(self._points > 0):
            d = np.diff(np.log(self._points))
            return np.allclose(d[0], d)
        return False

    @log.setter
    def log(self, log):
        if self._edges is not None:
            self._edges.log = log
        self._log = log

    @property
    def has_edges(self):
        '''True if edges are set'''
        return self._edges is not None and self._edges._edges is not None

    def __len__(self):
        if self._points is not None:
            return len(self._points)
        elif self._edges is not None:
            return len(self._edges)
        elif self._nbins is not None:
            return self._nbins
        return None

    def __str__(self):
        strs = []
        strs.append('(points) %s' % (self._points))
        strs.append('(edges)  %s' % (self._edges))
        strs.append('(nbins)  %s' % (self.nbins))
        return '\n'.join(strs)

    def __repr__(self):
        strs = []
        strs.append('Axis("%s",' % self.var)
        strs.append('points = %s,' % (self._points.__repr__()))
        strs.append('edges = %s)' % (self._edges.__repr__()))
        strs.append('nbins = %s)' % (self.nbins))
        return '\n'.join(strs)

    def __getitem__(self, idx):

        idx = self.convert_slice(idx)

        if idx is Ellipsis:
            return self

        new_obj = dm.Axis()
        new_obj.var = self.var
        if self.has_edges:
            new_obj._edges = self._edges[idx]
        if self._points is not None:
            new_obj._points = self._points[idx]
        new_obj._nbins = self._nbins
        return new_obj

    def convert_slice(self, idx):
        '''Convert slice

        idx : int, float, slice, Ellipsis
        '''
        if isinstance(idx, (int, np.integer, type(Ellipsis))):
            return idx

        if isinstance(idx, float):
            return self.convert_index(idx)

        if isinstance(idx, (list, np.ndarray)):
            new_indices = []
            for i in idx:
                new_indices.append(self.convert_index(i))
            return new_indices

        if isinstance(idx, slice):
            start = self.convert_index(idx.start)
            stop = self.convert_index(idx.stop)

            return slice(start, stop, idx.step)

        raise IndexError(idx, type(idx))

    def convert_index(self, idx):
        if idx is None:
            return None

        if isinstance(idx, (int, np.integer)):
            return idx

        idx = self.compute_indices(idx)
        if idx >= 0:
            return idx

        raise IndexError('Index out of range')

    def compute_indices(self, sample):
        '''compute bin indices for a sample, return -1 if outise of bins

        Parameters
        ----------
        sample : array, float

        Returns
        -------

        indices : array, int
        '''
        if not self.edges.consecutive:
            raise NotImplementedError()

        bins = self.edges.squeezed_edges
        if np.isscalar(sample):
            if sample == bins[-1]:
                return len(self)
            elif sample < bins[0] or sample > bins[-1]:
                return -1
            else:
                return np.digitize(sample, bins) - 1

        idx = np.digitize(sample, bins) - 1
        # make inclusive right edge
        idx[sample == bins[-1]] -= 1
        # set overflow bin to idx -1
        idx[idx == len(self)] = -1
        return idx

    @property
    def initialized(self):
        '''
        True if either edges or points are not None
        '''
        return self.has_edges or self.has_points

    def __eq__(self, other):
        if not type(self) == type(other):
            return False
        equal = self.var == other.var
        equal = equal and self._edges == other._edges
        equal = equal and np.all(np.equal(self._points, other._points))
        return equal and self._nbins == other._nbins

    @property
    def regular(self):
        '''True if spacing of egdges and/or points is regular'''
        regular = True
        if self._points is not None:
            if self.log:
                d = np.diff(np.log(self._points))
            else:
                d = np.diff(self._points)
            regular = regular and np.allclose(d[0], d)
        if self.has_edges:
            regular = regular and self._edges.regular
        return regular

    @property
    def edges(self):
        if self.has_edges:
            return self._edges
        if self.has_points:
            return dm.Edges(points=self._points, log=self.log)
        return None

    @edges.setter
    def edges(self, edges):
        edges = dm.Edges(edges)
        if self.initialized:
            if not len(edges) == len(self):
                raise IndexError('incompatible length of edges')
        self._edges = edges

    @property
    def points(self):
        if self.has_points:
            return self._points
        elif self.has_edges:
            return self._edges.points
        return None

    @points.setter
    def points(self, points):
        if self.initialized:
            if not len(points) == len(self):
                raise IndexError('incompatible length of points')
        self._points = points

    @property
    def squeezed_edges(self):
        return self.edges.squeezed_edges

    @property
    def nbins(self):
        if not (self.has_points and self.has_edges):
            return self._nbins
        else:
            return len(self.points)

    @nbins.setter
    def nbins(self, nbins):
        if not self.initialized:
            self._nbins = nbins
        else:
            raise ValueError('Cannot set n since bins are already defined')
